"""
This file contains several queries and logic to fetch statistics that can be displayed on the statistics views.
Please note that this file is only meant for the statistics app, and can change at any time!
"""
import asyncio

from datetime import datetime
from peewee import JOIN

from pyplanet.apps.contrib.local_records import LocalRecord
from pyplanet.apps.core.maniaplanet.models import Map, Player
from pyplanet.apps.core.statistics.models import Score, fn


class StatisticsProcessor:
	"""
	Statistics Processor.
	Please don't use this outside of the statistics app as the API will change without any info before.
	"""
	def __init__(self, app):
		"""
		Init the processor

		:param app: App instance.
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		self.app = app

		self.topsums_cache = None
		self.topsums_cache_time = None

	async def get_dashboard_data(self, player):
		"""
		Get player fact numbers, such as the number of finishes. Number of top-3 records, etc.

		:param player: Player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: dictionary with results.
		"""
		# Combine several calls.
		finishes, top_3, records = await asyncio.gather(
			self.get_num_finishes(player),
			self.get_num_top_3(player),
			self.get_num_records(player),
		)

		return dict(
			numbers=dict(
				finishes=finishes,
				top_3=top_3,
				records=records,
			),
		)

	async def get_num_finishes(self, player):
		"""
		Get the number of finishes.

		:param player: Player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: count or False when not possible to fetch.
		:rtype: int
		"""
		if self.app.instance.game.game == 'tm':
			return await Score.objects.count(
				Score.select(Score).where(Score.player == player)
			)
		return False

	async def get_num_records(self, player):
		"""
		Get the number of local records.

		:param player: Player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: count or False when not possible to fetch.
		:rtype: int
		"""
		if self.app.instance.game.game == 'tm' and 'local_records' in self.app.instance.apps.apps:
			return await LocalRecord.objects.count(
				LocalRecord.select(LocalRecord).where(LocalRecord.player == player)
			)
		return False

	async def get_num_top_3(self, player):
		"""
		Get the number of top-3 records of one player.

		:param player: Player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: count or False when not possible to fetch.
		:rtype: int
		"""
		# Get the players number of top-3 records (tm only + when local records is active).
		if self.app.instance.game.game == 'tm' and 'local_records' in self.app.instance.apps.apps:
			top = 0
			records = await LocalRecord.objects.execute(
				LocalRecord.select(LocalRecord).where(LocalRecord.player == player)
			)
			for record in records:
				top_3 = await LocalRecord.objects.execute(
					LocalRecord.select(LocalRecord).where(LocalRecord.map_id == record.map_id).limit(3)
				)
				if record in top_3:
					top += 1
			return top
		return False

	async def get_topsums(self):
		"""
		Get the topsums of the server.

		:return: List of top 100 players on the server with the statistics.
		"""
		if 'local_records' not in self.app.instance.apps.apps:
			return None

		now = datetime.now()
		diff = now - self.topsums_cache_time if self.topsums_cache_time else False
		if not self.topsums_cache or not diff or diff.total_seconds() > 60:
			self.topsums_cache = None
			self.topsums_cache_time = None
		elif self.topsums_cache:
			return self.topsums_cache

		maps = self.app.instance.map_manager.maps
		players = dict()

		for map_instance in maps:
			res = await LocalRecord.objects.execute(
				LocalRecord.select(LocalRecord, Player)
					.join(Player)
					.where(LocalRecord.map_id == map_instance.id)
					.order_by(LocalRecord.score)
					.limit(3)
			)
			for rank, entry in enumerate(res):
				if entry.player not in players:
					players[entry.player] = [0, 0, 0]
				players[entry.player][rank] += 1

		topsums = list(players.items())
		topsums.sort(key=lambda item: item[1][0] + item[1][1] + item[1][2], reverse=True)

		self.topsums_cache = topsums[:100]
		self.topsums_cache_time = datetime.now()
		return self.topsums_cache

	async def get_top_active_players(self):
		"""
		Get the top active players of the server.

		:return: List of top 100 players on the server ordered by the total playtimes.
		"""
		return await Player.objects.execute(
			Player.select(Player)
				.order_by(-Player.total_playtime)
				.limit(100)
		)

	async def get_top_donating_players(self):
		"""
		Get the top donating players of the server.

		:return: List of top 100 players on the server ordered by the total donations.
		"""
		return await Player.objects.execute(
			Player.select(Player)
				.order_by(-Player.total_donations)
				.limit(100)
		)



from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from ..models.mapfolder import MapFolder


def upgrade(migrator: SchemaMigrator):
	try:
		migrate(
			migrator.drop_index(MapFolder._meta.db_table, 'mapfolder_name'),
		)
	except Exception as e:
		print(str(e))
		print('Migration failed but silencing error!')


def downgrade(migrator: SchemaMigrator):
	pass



import asyncio
import platform
import aiohttp
import sys

from json import dumps

from pyplanet import __version__ as version
from pyplanet.conf import settings


class _Analytics:
	URL = 'https://api.amplitude.com/httpapi'
	KEY = '4a33dd84e315e0b02066a7993043a6c7'

	def __init__(self):
		self.client = None
		self.instance = None

	async def execute(self, *event):
		if not settings.ANALYTICS or settings.DEBUG:
			return

		await self.client.post(self.URL, params=dict(api_key=self.KEY, event=dumps(event)))

	async def capture(
		self, event_name, event_properties=None
	):
		event_properties = event_properties or dict()
		user_properties = dict(
			path=self.instance.game.server_path,
			game=self.instance.game.game,
			server_name=self.instance.game.server_name,
			language=self.instance.game.server_language,
			login=self.instance.game.server_player_login,
			title=self.instance.game.dedicated_title,
			dedicated_build=self.instance.game.dedicated_build,
			dedicated_version=self.instance.game.dedicated_version,
			python_version='{}.{}'.format(sys.version_info[0], sys.version_info[1]),
			python_version_full='{}.{}.{}'.format(*sys.version_info[0:3])
		)

		await self.execute(dict(
			user_id=self.instance.game.server_player_login,
			event_type=event_name, event_properties=event_properties, user_properties=user_properties,
			app_version=version, platform=platform.platform(), os_name=platform.system(), os_version=platform.version(),
			language=self.instance.game.server_language
		))

	async def start(self, instance):
		"""
		Initiate the analytics.

		:param instance: Instance of controller.
		:type instance: pyplanet.core.instance.Instance
		"""
		self.instance = instance
		self.client = aiohttp.ClientSession()

		try:
			await self.capture('start_controller', dict(
				app_labels=list(self.instance.apps.apps.keys())
			))
		except:
			pass

		# Start report loop
		asyncio.ensure_future(self.loop())

	async def loop(self):
		while True:
			await asyncio.sleep(60*5)
			try:
				await self.capture('online_ping', dict(
					total_players=self.instance.player_manager.count_all,
				))
			except:
				pass

Analytics = _Analytics()



import functools
import warnings


def deprecated(func):
	"""This is a decorator which can be used to mark functions
	as deprecated. It will result in a warning being emitted
	when the function is used.

	## Usage examples ##
	@deprecated
	def my_func():
		pass

	@other_decorators_must_be_upper
	@deprecated
	def my_func():
		pass
	"""

	@functools.wraps(func)
	def new_func(*args, **kwargs):
		warnings.warn_explicit(
			"Call to deprecated function %(funcname)s." % {
				'funcname': func.__name__,
			},
			category=DeprecationWarning,
			filename=func.__code__.co_filename,
			lineno=func.__code__.co_firstlineno + 1
		)
		return func(*args, **kwargs)

	return new_func



"""
Maniaplanet app component.
"""
from pyplanet.apps.core.statistics.views.players import TopActivePlayersView, TopDonatingPlayersView
from pyplanet.contrib.command import Command


class ManiaplanetComponent:
	def __init__(self, app):
		"""
		Initiate maniaplanet statistics component.

		:param app: App config instance
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		self.app = app

	async def on_init(self):
		pass

	async def on_start(self):
		# Register commands.
		await self.app.instance.command_manager.register(
			Command('topactive', target=self.top_active,
					description='Displays the most active players on this server.'),
			Command('topdons', target=self.top_donators,
					description='Displays the players who donated the most planets to this server.'),
		)

	async def top_active(self, player, *args, **kwargs):
		view = TopActivePlayersView(self.app, player, await self.app.processor.get_top_active_players())
		await view.display(player)

	async def top_donators(self, player, *args, **kwargs):
		view = TopDonatingPlayersView(self.app, player, await self.app.processor.get_top_donating_players())
		await view.display(player)



"""
Python helper for Semantic Versioning (http://semver.org/)

Original source: https://github.com/k-bx/python-semver
We copy to make sure this simple library is and maintain in our core.

Changed by Tom Valk: Added function to detect if the release is a pre-release.
"""

import collections
import re


__version__ = '2.7.7'
__author__ = 'Kostiantyn Rybnikov'
__author_email__ = 'k-bx@k-bx.com'

_REGEX = re.compile(
	r"""
	^
	(?P<major>(?:0|[1-9][0-9]*))
	\.
	(?P<minor>(?:0|[1-9][0-9]*))
	\.
	(?P<patch>(?:0|[1-9][0-9]*))
	(\-*(?P<prerelease>
		(?:0|[1-9A-Za-z-][0-9A-Za-z-]*)
		(\.(?:0|[1-9A-Za-z-][0-9A-Za-z-]*))*
	))?
	(\+(?P<build>
		[0-9A-Za-z-]+
		(\.[0-9A-Za-z-]+)*
	))?
	$
	""", re.VERBOSE)

_LAST_NUMBER = re.compile(r'(?:[^\d]*(\d+)[^\d]*)+')

if not hasattr(__builtins__, 'cmp'):
	def cmp(a, b):
		return (a > b) - (a < b)


def parse(version):
	"""Parse version to major, minor, patch, pre-release, build parts.

	:param version: version string
	:return: dictionary with the keys 'build', 'major', 'minor', 'patch',
			 and 'prerelease'. The prerelease or build keys can be None
			 if not provided
	:rtype: dict
	"""
	match = _REGEX.match(version)
	if match is None:
		raise ValueError('%s is not valid SemVer string' % version)

	version_parts = match.groupdict()

	version_parts['major'] = int(version_parts['major'])
	version_parts['minor'] = int(version_parts['minor'])
	version_parts['patch'] = int(version_parts['patch'])

	return version_parts


class VersionInfo(collections.namedtuple(
	'VersionInfo', 'major minor patch prerelease build')):
	"""
	:param int major: version when you make incompatible API changes.
	:param int minor: version when you add functionality in
					  a backwards-compatible manner.
	:param int patch: version when you make backwards-compatible bug fixes.
	:param str prerelease: an optional prerelease string
	:param str build: an optional build string

	>>> import semver
	>>> ver = semver.parse('3.4.5-pre.2+build.4')
	>>> ver
	{'build': 'build.4', 'major': 3, 'minor': 4, 'patch': 5,
	'prerelease': 'pre.2'}
	"""
	__slots__ = ()

	def __eq__(self, other):
		if not isinstance(other, (VersionInfo, dict)):
			return NotImplemented
		return _compare_by_keys(self._asdict(), _to_dict(other)) == 0

	def __ne__(self, other):
		if not isinstance(other, (VersionInfo, dict)):
			return NotImplemented
		return _compare_by_keys(self._asdict(), _to_dict(other)) != 0

	def __lt__(self, other):
		if not isinstance(other, (VersionInfo, dict)):
			return NotImplemented
		return _compare_by_keys(self._asdict(), _to_dict(other)) < 0

	def __le__(self, other):
		if not isinstance(other, (VersionInfo, dict)):
			return NotImplemented
		return _compare_by_keys(self._asdict(), _to_dict(other)) <= 0

	def __gt__(self, other):
		if not isinstance(other, (VersionInfo, dict)):
			return NotImplemented
		return _compare_by_keys(self._asdict(), _to_dict(other)) > 0

	def __ge__(self, other):
		if not isinstance(other, (VersionInfo, dict)):
			return NotImplemented
		return _compare_by_keys(self._asdict(), _to_dict(other)) >= 0


def _to_dict(obj):
	if isinstance(obj, VersionInfo):
		return obj._asdict()
	return obj


def parse_version_info(version):
	"""Parse version string to a VersionInfo instance.

	:param version: version string
	:return: a :class:`VersionInfo` instance
	:rtype: :class:`VersionInfo`
	"""
	parts = parse(version)
	version_info = VersionInfo(
		parts['major'], parts['minor'], parts['patch'],
		parts['prerelease'], parts['build'])

	return version_info


def _nat_cmp(a, b):
	def convert(text):
		return int(text) if re.match('[0-9]+', text) else text

	def split_key(key):
		return [convert(c) for c in key.split('.')]

	def cmp_prerelease_tag(a, b):
		if isinstance(a, int) and isinstance(b, int):
			return cmp(a, b)
		elif isinstance(a, int):
			return -1
		elif isinstance(b, int):
			return 1
		else:
			return cmp(a, b)

	a, b = a or '', b or ''
	a_parts, b_parts = split_key(a), split_key(b)
	for sub_a, sub_b in zip(a_parts, b_parts):
		cmp_result = cmp_prerelease_tag(sub_a, sub_b)
		if cmp_result != 0:
			return cmp_result
	else:
		return cmp(len(a), len(b))


def _compare_by_keys(d1, d2):
	for key in ['major', 'minor', 'patch']:
		v = cmp(d1.get(key), d2.get(key))
		if v:
			return v

	rc1, rc2 = d1.get('prerelease'), d2.get('prerelease')
	rccmp = _nat_cmp(rc1, rc2)

	if not rccmp:
		return 0
	if not rc1:
		return 1
	elif not rc2:
		return -1

	return rccmp


def compare(ver1, ver2):
	"""Compare two versions

	:param ver1: version string 1
	:param ver2: version string 2
	:return: The return value is negative if ver1 < ver2,
			 zero if ver1 == ver2 and strictly positive if ver1 > ver2
	:rtype: int
	"""

	v1, v2 = parse(ver1), parse(ver2)

	return _compare_by_keys(v1, v2)


def match(version, match_expr):
	"""Compare two versions through a comparison

	:param str version: a version string
	:param str match_expr: operator and version; valid operators are
		  <   smaller than
		  >   greater than
		  >=  greator or equal than
		  <=  smaller or equal than
		  ==  equal
		  !=  not equal
	:return: True if the expression matches the version, otherwise False
	:rtype: bool
	"""
	prefix = match_expr[:2]
	if prefix in ('>=', '<=', '==', '!='):
		match_version = match_expr[2:]
	elif prefix and prefix[0] in ('>', '<'):
		prefix = prefix[0]
		match_version = match_expr[1:]
	else:
		raise ValueError("match_expr parameter should be in format <op><ver>, "
						 "where <op> is one of "
						 "['<', '>', '==', '<=', '>=', '!=']. "
						 "You provided: %r" % match_expr)

	possibilities_dict = {
		'>': (1,),
		'<': (-1,),
		'==': (0,),
		'!=': (-1, 1),
		'>=': (0, 1),
		'<=': (-1, 0)
	}

	possibilities = possibilities_dict[prefix]
	cmp_res = compare(version, match_version)

	return cmp_res in possibilities


def max_ver(ver1, ver2):
	"""Returns the greater version of two versions

	:param ver1: version string 1
	:param ver2: version string 2
	:return: the greater version of the two
	:rtype: :class:`VersionInfo`
	"""
	cmp_res = compare(ver1, ver2)
	if cmp_res == 0 or cmp_res == 1:
		return ver1
	else:
		return ver2


def min_ver(ver1, ver2):
	"""Returns the smaller version of two versions

	:param ver1: version string 1
	:param ver2: version string 2
	:return: the smaller version of the two
	:rtype: :class:`VersionInfo`
	"""
	cmp_res = compare(ver1, ver2)
	if cmp_res == 0 or cmp_res == -1:
		return ver1
	else:
		return ver2


def format_version(major, minor, patch, prerelease=None, build=None):
	"""Format a version according to the Semantic Versioning specification

	:param str major: the required major part of a version
	:param str minor: the required minor part of a version
	:param str patch: the required patch part of a version
	:param str prerelease: the optional prerelease part of a version
	:param str build: the optional build part of a version
	:return: the formatted string
	:rtype: str
	"""
	version = "%d.%d.%d" % (major, minor, patch)
	if prerelease is not None:
		version = version + "-%s" % prerelease

	if build is not None:
		version = version + "+%s" % build

	return version


def _increment_string(string):
	"""
	Look for the last sequence of number(s) in a string and increment, from:
	http://code.activestate.com/recipes/442460-increment-numbers-in-a-string/#c1
	"""
	match = _LAST_NUMBER.search(string)
	if match:
		next_ = str(int(match.group(1)) + 1)
		start, end = match.span(1)
		string = string[:max(end - len(next_), start)] + next_ + string[end:]
	return string


def bump_major(version):
	"""Raise the major part of the version

	:param: version string
	:return: the raised version string
	:rtype: str
	"""
	verinfo = parse(version)
	return format_version(verinfo['major'] + 1, 0, 0)


def bump_minor(version):
	"""Raise the minor part of the version

	:param: version string
	:return: the raised version string
	:rtype: str
	"""
	verinfo = parse(version)
	return format_version(verinfo['major'], verinfo['minor'] + 1, 0)


def bump_patch(version):
	"""Raise the patch part of the version

	:param: version string
	:return: the raised version string
	:rtype: str
	"""
	verinfo = parse(version)
	return format_version(verinfo['major'], verinfo['minor'],
						  verinfo['patch'] + 1)


def bump_prerelease(version):
	"""Raise the prerelease part of the version

	:param: version string
	:return: the raised version string
	:rtype: str
	"""
	verinfo = parse(version)
	verinfo['prerelease'] = _increment_string(verinfo['prerelease'] or 'rc.0')
	return format_version(verinfo['major'], verinfo['minor'], verinfo['patch'],
						  verinfo['prerelease'])


def bump_build(version):
	"""Raise the build part of the version

	:param: version string
	:return: the raised version string
	:rtype: str
	"""
	verinfo = parse(version)
	verinfo['build'] = _increment_string(verinfo['build'] or 'build.0')
	return format_version(verinfo['major'], verinfo['minor'], verinfo['patch'],
						  verinfo['prerelease'], verinfo['build'])


def is_prerelease(version):
	"""
	Detect if the version given is a pre-release.

	:param version: version string
	:return: if the version is a pre-release
	:rtype: bool
	"""
	verinfo = parse(version)
	return 'prerelease' in verinfo and verinfo['prerelease'] is not None



import os

from jinja2 import PrefixLoader, FileSystemLoader, PackageLoader


class _PyPlanetLoader(PrefixLoader):

	def __init__(self, mapping=None, delimiter='/'):
		super().__init__(mapping=mapping or self.get_mapping(), delimiter=delimiter)

	@classmethod
	def get_mapping(cls):
		from pyplanet.core import Controller

		mapping = dict()

		# Static core components.
		mapping['core.views'] = PackageLoader('pyplanet.views', 'templates')

		# Add app prefixes.
		for app_label, app in Controller.instance.apps.apps.items():
			template_path = os.path.join(app.path, 'templates')
			if os.path.exists(template_path):
				mapping[app_label] = FileSystemLoader(template_path)

		return mapping


class PyPlanetLoader:
	"""
	Lazy loader for the pyplanet jinja2 loader.
	"""
	_INSTANCE = None

	@classmethod
	def get_loader(cls):
		if not cls._INSTANCE:
			cls._INSTANCE = _PyPlanetLoader()
		return cls._INSTANCE



"""
Mode Settings Views.
"""

from pyplanet.views.generics import ManualListView


class PlayerListView(ManualListView):
	title = 'Players'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Buddies'

	def __init__(self, app, player):
		"""
		:param app: App config instance.
		:param player: Player instance.
		:type app: pyplanet.apps.contrib.admin.Admin
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		"""
		super().__init__()
		self.manager = app.context.ui
		self.app = app
		self.player = player
		self.fields = [
			{
				'name': 'Nickname',
				'index': 'nickname',
				'sorting': True,
				'searching': True,
				'width': 60,
				'type': 'label'
			},
			{
				'name': 'Login',
				'index': 'login',
				'input': True,
				'sorting': False,
				'searching': True,
				'width': 40,
				'type': 'label',
			},
			{
				'name': 'Spec',
				'index': 'is_spectator',
				'sorting': True,
				'searching': False,
				'width': 15,
				'type': 'label',
				'safe': True
			},
			{
				'name': 'Level',
				'index': 'level',
				'sorting': True,
				'searching': False,
				'width': 35,
				'type': 'label'
			},
		]
		self.sort_field = self.fields[2]

		self.child = None

	async def get_data(self):
		players = self.app.instance.player_manager.online
		return [dict(
			nickname=p.nickname,
			login='{}'.format(p.login),
			is_spectator='$f00&#xf03d;' if p.flow.is_spectator else '$73f&#xf007;',
			is_spectator_bool=p.flow.is_spectator,
			level='{}: {}'.format(p.level, p.get_level_string())
		) for p in players]

	async def display(self, **kwargs):
		kwargs['player'] = self.player
		return await super().display(**kwargs)

	async def get_actions(self):
		return [
			{
				'name': 'Force spec/player',
				'type': 'label',
				'text': 'Force',
				'width': 12,
				'action': self.action_force,
				'safe': True,
			},
			{
				'name': 'Ignore',
				'type': 'label',
				'text': 'Ignore',
				'width': 12,
				'action': self.action_ignore,
				'safe': True,
			},
			{
				'name': 'Warn',
				'type': 'label',
				'text': 'Warn',
				'width': 12,
				'action': self.action_warn,
				'safe': True,
			},
			{
				'name': 'Kick',
				'type': 'label',
				'text': 'Kick',
				'width': 12,
				'action': self.action_kick,
				'safe': True,
			},
			{
				'name': 'Ban',
				'type': 'label',
				'text': 'Ban',
				'width': 12,
				'action': self.action_ban,
				'safe': True,
			},
			{
				'name': 'Blacklist',
				'type': 'label',
				'text': 'Blacklist',
				'width': 12,
				'action': self.action_blacklist,
				'safe': True,
			},
		]

	async def action_force(self, user, values, player, *args, **kwargs):
		if player['is_spectator_bool']:
			await self.app.instance.command_manager.execute(user, '//forceplay', player['login'])
		else:
			await self.app.instance.command_manager.execute(user, '//forcespec', player['login'])
		await self.refresh(self.player)

	async def action_ignore(self, user, values, player, *args, **kwargs):
		ignore_list = await self.app.instance.gbx('GetIgnoreList', -1, 0)
		in_list = False
		for entry in ignore_list:
			if entry and entry['Login'] == player['login']:
				in_list = True
				break

		if not in_list:
			await self.app.instance.command_manager.execute(user, '//ignore', player['login'])
		else:
			await self.app.instance.command_manager.execute(user, '//unignore', player['login'])
		await self.refresh(self.player)

	async def action_warn(self, user, values, player, *args, **kwargs):
		await self.app.instance.command_manager.execute(user, '//warn', player['login'])

	async def action_kick(self, user, values, player, *args, **kwargs):
		await self.app.instance.command_manager.execute(user, '//kick', player['login'])
		await self.refresh(self.player)

	async def action_ban(self, user, values, player, *args, **kwargs):
		await self.app.instance.command_manager.execute(user, '//ban', player['login'])
		await self.refresh(self.player)

	async def action_blacklist(self, user, values, player, *args, **kwargs):
		await self.app.instance.command_manager.execute(user, '//blacklist', player['login'])
		await self.refresh(self.player)



import logging

from pyplanet.contrib import CoreContrib
from pyplanet.contrib.mode.signals import script_mode_changed

logger = logging.getLogger(__name__)


class ModeManager(CoreContrib):
	"""
	Mode Manager manges the script, script settings and the mode UI settings of the current game mode.

	.. warning::

		Don't initiate this class yourself. Use ``instance.mode_manager`` for an static instance.

	"""
	def __init__(self, instance):
		"""
		Initiate, should only be done from the core instance.

		:param instance: Instance.
		:type instance: pyplanet.core.instance.Instance
		"""
		self._instance = instance

		self._current_script = None
		self._next_script = None
		self._current_full_script = None
		self._next_full_script = None

		self._next_settings_update = dict()
		self._next_variables_update = dict()

	async def on_start(self):
		"""
		Handle startup, just before the apps will start. We will make sure we are ready to get requests for permissions.
		"""
		self._current_script = await self.get_current_script(refresh=True)

		# Listeners.
		self._instance.signals.listen('maniaplanet:server_start', self._on_change)

	async def _on_change(self, *args, **kwargs):
		# Making sure we set the settings + variables.
		if len(self._next_settings_update.keys()) > 0:
			logger.debug('Setting mode settings right now!')
			try:
				await self.update_settings(self._next_settings_update)
			except Exception as e:
				logging.error('Can\'t set the script mode settings! Error: {}'.format(str(e)))
			self._next_settings_update = dict()
		if len(self._next_variables_update.keys()) > 0:
			logger.debug('Setting mode variables right now!')
			try:
				await self.update_variables(self._next_variables_update)
			except Exception as e:
				logging.error('Can\'t set the script mode variables! Error: {}'.format(str(e)))
			self._next_variables_update = dict()

		# Make sure we send to the signal when mode is been changed.
		if self._current_script != self._next_script:
			await script_mode_changed.send_robust({
				'unloaded_script': self._current_script, 'loaded_script': self._next_script
			})

		await self.get_current_script(refresh=True)

	async def get_current_script(self, refresh=False):
		"""
		Get the current script name.

		:param refresh: Refresh from server.
		"""
		if refresh or not self._current_script:
			payload = await self._instance.gbx('GetScriptName')
			current_value = payload['CurrentValue'].partition('.')[0]
			if '\\' in current_value:
				current_value = current_value.rpartition('\\')[2]
			self._current_script = current_value
			self._current_full_script = payload['CurrentValue']

			if 'NextValue' in payload:
				next_value = payload['NextValue'].partition('.')[0]
				if '\\' in next_value:
					next_value = next_value.rpartition('\\')[2]
				self._next_script = next_value
				self._next_full_script = payload['NextValue']

		return self._current_script

	async def get_next_script(self, refresh=False):
		"""
		Get the next script name.

		:param refresh: Refresh from server.
		"""
		await self.get_current_script(refresh=refresh)
		return self._next_script

	async def get_current_full_script(self, refresh=False):
		"""
		Get the current full script name.

		:param refresh: Refresh from server.
		"""
		if refresh or not self._current_full_script:
			await self.get_current_script(True)

		return self._current_full_script

	async def get_next_full_script(self, refresh=False):
		"""
		Get the next full script name.

		:param refresh: Refresh from server.
		"""
		if refresh or not self._next_full_script:
			await self.get_current_script(True)

		return self._next_full_script

	async def get_current_script_info(self):
		"""
		Get the script info as a structure containing: Name, CompatibleTypes, Description, Version and the settings available.
		"""
		return await self._instance.gbx('GetModeScriptInfo')

	async def set_next_script(self, name):
		"""
		Set the next played script name (after map restart/skip).

		:param name: Name
		"""
		await self._instance.gbx('SetScriptName', name)
		name = name.partition('.')[0]
		if '\\' in name:
			name = name.rpartition('\\')[2]
		self._next_script = name

	async def get_settings(self):
		"""
		Get the current mode settings as a dictionary.
		"""
		return await self._instance.gbx('GetModeScriptSettings')

	async def update_settings(self, update_dict):
		"""
		Update the current settings, merges current settings with the provided settings. Replaces by the keys you give
		if the data already exists.

		:param update_dict: The dictionary with the partial updated keys and values.
		"""
		current_settings = await self.get_settings()
		current_settings.update(update_dict)
		await self._instance.gbx('SetModeScriptSettings', current_settings)

	async def update_next_settings(self, update_dict):
		"""
		Queue setting changes for the next script (that will be active after restart).

		:param update_dict: The dictionary with the partial updated keys and values.
		"""
		if not isinstance(self._next_settings_update, dict):
			self._next_settings_update = dict()
		self._next_settings_update.update(update_dict)

	async def get_variables(self):
		"""
		Get the mode script variables.
		"""
		return await self._instance.gbx('GetModeScriptVariables')

	async def update_variables(self, update_dict):
		"""
		Update the current variables, merges current vars with the provided vars. Replaces by the keys you give
		if the data already exists.

		:param update_dict: The dictionary with the partial updated keys and values.
		"""
		variables = await self.get_variables()
		variables.update(update_dict)
		await self._instance.gbx('SetModeScriptVariables', variables)

	async def update_next_variables(self, update_dict):
		"""
		Queue variable changes for the next script (that will be active after restart).

		:param update_dict: The dictionary with the partial updated keys and values.
		"""
		if not isinstance(self._next_variables_update, dict):
			self._next_variables_update = dict()
		self._next_variables_update.update(update_dict)



"""
The list module contains the class to hold all the queued players.
"""

import asyncio
from random import shuffle

from pyplanet.apps.contrib.queue.exception import PlayerAlreadyInQueue


class QueueList:
	"""
	The Queue List holds the actual queue and manages it's list thread-safe.
	"""

	def __init__(self):
		self.list = list()
		self._lock = asyncio.Lock()
		self.lock = asyncio.Lock()

		self.change_hook = None

	async def push(self, player):
		"""
		Push a player in the last place of the queue. Return the place of the player.

		:param player: Player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: Returns the index of the player (the textual position is index+1).
		:rtype: int
		"""
		with await self._lock:
			# Make sure the player is not yet in the queue.
			try:
				index = self.list.index(player)
				raise PlayerAlreadyInQueue('Player {} already in queue on position {} (index: {})'.format(player.login, index+1, index))
			except ValueError:
				pass

			# Add the player to the queue.
			self.list.append(player)
			index = self.list.index(player)

		# Call change hook.
		if self.change_hook and callable(self.change_hook):
			await self.change_hook(action='push', entity=player)

		return index

	async def count(self):
		"""
		Get the current count of the queue.

		:return: Count of list.
		:rtype: int
		"""
		with await self._lock:
			return len(self.list)

	async def has(self, player):
		"""
		Check if the player is in the queue list.

		:param player: Player object
		:return: boolean
		:rtype: bool
		"""
		with await self._lock:
			return self.list.count(player) != 0

	async def get_position(self, player):
		"""
		Get the position of the player.

		:param player: Player instance
		:return: Integer of the position (index + 1) or None if not found.
		"""
		try:
			with await self._lock:
				return self.list.index(player) + 1
		except ValueError:
			return None

	async def copy(self):
		"""
		Copy the list and return the copy.

		:return: Copy of the player list.
		"""
		with await self._lock:
			return self.list.copy()

	async def pop(self):
		"""
		Get the next player and remove from the queue. Returns None when no next player is available (and the queue is empty).

		:return: Player object or None
		"""
		with await self._lock:
			if len(self.list) > 0:
				player = self.list.pop(0)
			else:
				return None

		# Call change hook.
		if self.change_hook and callable(self.change_hook):
			await self.change_hook(action='pop', entity=player)

		return player

	async def remove(self, player):
		"""
		Remove player from the queue, regardless of the position.

		:param player: Player instance
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: Returns if entry has been removed (bool)
		:rtype: bool
		"""
		with await self._lock:
			try:
				self.list.remove(player)
			except ValueError:
				return False

		# Call change hook.
		if self.change_hook and callable(self.change_hook):
			await self.change_hook(action='remove', entity=player)

		return True

	async def clear(self):
		"""
		Clear and remove everyone from the queue.
		"""
		with await self._lock:
			self.list.clear()

		# Call change hook.
		if self.change_hook and callable(self.change_hook):
			await self.change_hook(action='clear', entity=None)

	async def shuffle(self):
		"""
		Shuffle the queue.
		"""
		with await self._lock:
			shuffle(self.list)

		# Call change hook.
		if self.change_hook and callable(self.change_hook):
			await self.change_hook(action='shuffle', entity=None)



import math

from pyplanet.utils.style import style_strip
from pyplanet.utils.times import format_time
from pyplanet.views.generics.widget import TimesWidgetView
from pyplanet.views.generics.list import ManualListView
from pyplanet.utils import times


class DedimaniaRecordsWidget(TimesWidgetView):
	widget_x = -160
	widget_y = 70.5
	z_index = 130
	size_x = 38
	size_y = 55.5
	top_entries = 5
	title = 'Dedimania'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widgets_dedimaniarecords'

		self.action = self.action_recordlist
		self.record_amount = 15

	async def get_player_data(self):
		data = await super().get_player_data()
		if self.app.instance.performance_mode:
			return data

		widget_times = dict()

		for player in self.app.instance.player_manager.online:
			player_record = [x for x in self.app.current_records if x.login == player.login]
			player_index = (len(self.app.current_records) + 1)
			list_records = list()
			if len(player_record) > 0:
				# Set player index if there is a record
				player_index = (self.app.current_records.index(player_record[0]) + 1)

			records = list(self.app.current_records[:self.top_entries])
			custom_start_index = None
			if player_index > len(self.app.current_records):
				# No personal record, get the last records
				records_start = (len(self.app.current_records) - self.record_amount + self.top_entries)
				# If start of current slice is in the top entries, add more records below
				if records_start < self.top_entries:
					records_start = self.top_entries

				records += list(self.app.current_records[records_start:])
				custom_start_index = (records_start + 1)
			else:
				if player_index <= self.top_entries:
					# Player record is in top X, get following records (top entries + 1 onwards)
					records += self.app.current_records[self.top_entries:self.record_amount]
					custom_start_index = (self.top_entries + 1)
				else:
					# Player record is not in top X, get records around player record
					# Same amount above the record as below, except when not possible (favors above)
					records_to_fill = (self.record_amount - self.top_entries)
					start_point = ((player_index - math.ceil((records_to_fill - 1) / 2)) - 1)
					end_point = ((player_index + math.floor((records_to_fill - 1) / 2)) - 1)

					# If end of current slice is outside the list, add more records above
					if end_point > len(self.app.current_records):
						end_difference = (end_point - len(self.app.current_records))
						start_point = (start_point - end_difference)
					# If start of current slice is in the top entries, add more records below
					if start_point < self.top_entries:
						start_point = self.top_entries

					records += self.app.current_records[start_point:(start_point + records_to_fill)]
					custom_start_index = (start_point + 1)

			index = 1
			for record in records:
				list_record = dict()
				list_record['index'] = index
				list_record['color'] = '$fff'
				if index <= self.top_entries:
					list_record['color'] = '$ff0'
				if index == player_index:
					list_record['color'] = '$0f3'
				list_record['nickname'] = record.nickname
				list_record['score'] = times.format_time(int(record.score))
				if index == self.top_entries:
					index = custom_start_index
				else:
					index += 1
				list_records.append(list_record)

			widget_times[player.login] = {'times': list_records}

		data.update(widget_times)

		return data

	async def get_context_data(self):
		context = await super().get_context_data()

		# Add facts.
		context.update({
			'top_entries': self.top_entries
		})

		if self.app.instance.performance_mode:
			list_records = list()
			records = list(self.app.current_records[:self.record_amount])

			index = 1
			for record in records:
				list_record = dict()
				list_record['index'] = index
				list_record['color'] = '$fff'
				if index <= self.top_entries:
					list_record['color'] = '$ff0'
				list_record['nickname'] = record.nickname
				list_record['score'] = times.format_time(int(record.score))
				index += 1
				list_records.append(list_record)

			context.update({
				'times': list_records
			})

		return context

	async def action_recordlist(self, player, **kwargs):
		await self.app.show_records_list(player)


class DedimaniaRecordsListView(ManualListView):
	title = 'Dedimania Records on this map'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	fields = [
		{
			'name': '#',
			'index': 'index',
			'sorting': True,
			'searching': False,
			'width': 10,
			'type': 'label'
		},
		{
			'name': 'Player',
			'index': 'nickname',
			'sorting': False,
			'searching': True,
			'width': 70
		},
		{
			'name': 'Time',
			'index': 'record_time',
			'sorting': True,
			'searching': False,
			'width': 30,
			'type': 'label'
		},
		{
			'name': 'Difference',
			'index': 'record_time_difference',
			'sorting': True,
			'searching': False,
			'width': 50,
			'type': 'label'
		},
	]

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui

	async def get_title(self):
		return 'Dedimania Records on $l[http://dedimania.net/tm2stats/?do=stat&UId={}&Show=RECORDS]{}$l'.format(
			self.app.instance.map_manager.current_map.uid, self.app.instance.map_manager.current_map.name
		)

	async def get_data(self):
		index = 1
		items = []
		async with self.app.lock:
			first_time = self.app.current_records[0].score
			for item in self.app.current_records:
				record_time_difference = ''
				if index > 1:
					record_time_difference = '$f00 + ' + times.format_time((item.score - first_time))
				items.append({
					'index': index, 'nickname': item.nickname,
					'record_time': times.format_time(item.score),
					'record_time_difference': record_time_difference
				})
				index += 1

		return items


class DedimaniaCpCompareListView(ManualListView):
	title = 'Dedimania checkpoint comparison'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app, own_record, compare_record):
		"""
		Init compare listview.

		:param app: App instance
		:param own_record: Own record
		:param compare_record: Compare with record.
		:type own_record: pyplanet.apps.contrib.dedimania.api.DedimaniaRecord
		:type compare_record: pyplanet.apps.contrib.dedimania.api.DedimaniaRecord
		"""
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.provide_search = False

		self.own_record = own_record
		self.compare_record = compare_record

	async def get_fields(self):
		return [
			{
				'name': 'Checkpoint',
				'index': 'cp',
				'sorting': False,
				'searching': False,
				'width': 40,
				'type': 'label'
			},
			{
				'name': '#{}: $n{}'.format(self.own_record.rank, style_strip(self.own_record.nickname)) if self.own_record else '-',
				'index': 'own_time',
				'sorting': False,
				'searching': False,
				'width': 70
			},
			{
				'name': '#{}: $n{}'.format(self.compare_record.rank, style_strip(self.compare_record.nickname)),
				'index': 'compare_time',
				'sorting': False,
				'searching': False,
				'width': 70,
			},
			{
				'name': 'Difference',
				'index': 'difference',
				'sorting': False,
				'searching': False,
				'width': 50,
				'type': 'label'
			},
		]

	async def get_title(self):
		return 'Dedimania CP comparison on {}'.format(self.app.instance.map_manager.current_map.name)

	def get_diff_text(self, a, b):
		diff = a - b
		diff_prefix = '$FFF'
		if diff > 0:
			diff_prefix = '$F66+'  # Red
		elif diff < 0:
			diff_prefix = '$6CF- '  # Blue

		return '{}{}'.format(diff_prefix, format_time(abs(diff)))

	async def get_data(self):
		compare_cps = self.compare_record.cps
		own_cps = self.own_record.cps if self.own_record else [None for _ in compare_cps]
		total_cps = len(own_cps)

		data = list()
		for cp, (own, compare) in enumerate(zip(own_cps, compare_cps)):
			data.append(dict(
				cp='Finish' if (cp + 1) == total_cps else cp + 1,
				own_time=format_time(own) if own else '',
				compare_time=format_time(compare),
				difference=self.get_diff_text(own, compare) if self.own_record else '-'
			))

		return data





class MapException(Exception):
	"""Generic map exception by manager."""


class MapNotFound(MapException):
	"""Map not found"""


class MapIncompatible(MapException):
	"""The map you want to add/insert/upload is invalid and not suited for the current server config."""


class ModeIncompatible(MapException):
	"""The current mode doesn't support the given action."""



"""
This file contains the apps & apps settings and overrides the default ones that are defined in the core.
Please copy the default file to: apps.py
"""

# In the apps dictionary and array you configure the apps (or plugins) are loaded for specific pools (controllers).
# Be aware that the list will *ALWAYS* be prepended after the mandatory defaults are loaded in place.
# The mandatory defaults are specific per version, refer to the documentation:
APPS = {
	'default': [
		'pyplanet.apps.contrib.admin',
		'pyplanet.apps.contrib.jukebox',
		'pyplanet.apps.contrib.karma',
		'pyplanet.apps.contrib.local_records',
		'pyplanet.apps.contrib.dedimania',
		'pyplanet.apps.contrib.players',
		'pyplanet.apps.contrib.info',
		'pyplanet.apps.contrib.mx',
		'pyplanet.apps.contrib.transactions',

		# New since 0.4.0:
		'pyplanet.apps.contrib.sector_times',
		'pyplanet.apps.contrib.dynamic_points',

		# New since 0.5.0:
		'pyplanet.apps.contrib.clock',
		'pyplanet.apps.contrib.best_cps',
		'pyplanet.apps.contrib.voting',

		# New since 0.6.0:
		'pyplanet.apps.contrib.queue',
		'pyplanet.apps.contrib.ads',
		'pyplanet.apps.contrib.music_server',

		# New since 0.8.0:
		'pyplanet.apps.contrib.funcmd',

		# Local
		'tests.apps.migration_test.app.MigrationTest',
		'tests.apps.new_import_style',
	],
}



from pyplanet.views.generics.widget import WidgetView


class ClockWidget(WidgetView):
    widget_x = 150
    widget_y = -59
    template_name = 'clock/clock.xml'

    def __init__(self, app):
        super().__init__(self)
        self.app = app
        self.manager = app.context.ui
        self.id = 'pyplanet__widgets_clock'



from xmlrpc.client import Fault


class DedimaniaException(Exception):
	"""
	General exception for dedimania.
	"""
	pass


class DedimaniaTransportException(DedimaniaException):
	"""
	Transport, decode or encode issue.
	"""
	pass


class DedimaniaNotSupportedException(DedimaniaException):
	"""
	Map or mode is not supported by Dedimania.
	"""
	pass


class DedimaniaInvalidCredentials(DedimaniaException):
	"""
	Invalid code or player.
	"""
	pass


class DedimaniaFault(DedimaniaException, Fault):
	"""
	Dedimania XMLRPC Fault.
	"""
	pass



"""
AppVeyor initiate script

This is a port of https://github.com/pypa/python-packaging-user-guide/blob/master/source/code/install.ps1
with various fixes and improvements that just weren't feasible to implement in PowerShell.
"""
from __future__ import print_function
from os import environ
from os.path import exists
from subprocess import check_call

try:
	from urllib.request import urlretrieve
except ImportError:
	from urllib import urlretrieve

BASE_URL = "https://www.python.org/ftp/python/"
GET_PIP_URL = "https://bootstrap.pypa.io/get-pip.py"
GET_PIP_PATH = "C:\get-pip.py"
URLS = {
	# NOTE: no .msi installer for 3.3.6
	("3.5", "64"): BASE_URL + "3.5.4/python-3.5.4-amd64.exe",
	("3.5", "32"): BASE_URL + "3.5.4/python-3.5.4.exe",
	("3.6", "64"): BASE_URL + "3.6.8/python-3.6.8-amd64.exe",
	("3.6", "32"): BASE_URL + "3.6.8/python-3.6.8.exe",
	("3.7", "64"): BASE_URL + "3.7.7/python-3.7.7-amd64.exe",
	("3.7", "32"): BASE_URL + "3.7.7/python-3.7.7.exe",
	("3.8", "64"): BASE_URL + "3.8.2/python-3.8.2-amd64.exe",
	("3.8", "32"): BASE_URL + "3.8.2/python-3.8.2.exe",
}
INSTALL_CMD = {
	# Commands are allowed to fail only if they are not the last command.  Eg: uninstall (/x) allowed to fail.
	"2.6": [["msiexec.exe", "/L*+!", "install.log", "/qn", "/x", "{path}"],
			["msiexec.exe", "/L*+!", "install.log", "/qn", "/i", "{path}", "TARGETDIR={home}"]],
	"2.7": [["msiexec.exe", "/L*+!", "install.log", "/qn", "/x", "{path}"],
			["msiexec.exe", "/L*+!", "install.log", "/qn", "/i", "{path}", "TARGETDIR={home}"]],
	"3.3": [["msiexec.exe", "/L*+!", "install.log", "/qn", "/x", "{path}"],
			["msiexec.exe", "/L*+!", "install.log", "/qn", "/i", "{path}", "TARGETDIR={home}"]],
	"3.4": [["msiexec.exe", "/L*+!", "install.log", "/qn", "/x", "{path}"],
			["msiexec.exe", "/L*+!", "install.log", "/qn", "/i", "{path}", "TARGETDIR={home}"]],
	"3.5": [["{path}", "/quiet", "TargetDir={home}"]],
	"3.6": [["{path}", "/quiet", "TargetDir={home}"]],
	"3.7": [["{path}", "/quiet", "TargetDir={home}"]],
	"3.8": [["{path}", "/quiet", "TargetDir={home}"]],
}


def download_file(url, path):
	print("Downloading: {} (into {})".format(url, path))
	progress = [0, 0]

	def report(count, size, total):
		progress[0] = count * size
		if progress[0] - progress[1] > 1000000:
			progress[1] = progress[0]
			print("Downloaded {:,}/{:,} ...".format(progress[1], total))

	dest, _ = urlretrieve(url, path, reporthook=report)
	return dest


def install_python(version, arch, home):
	print("Installing Python", version, "for", arch, "bit architecture to", home)
	if exists(home):
		return

	path = download_python(version, arch)
	print("Installing", path, "to", home)
	success = False
	for cmd in INSTALL_CMD[version]:
		cmd = [part.format(home=home, path=path) for part in cmd]
		print("Running:", " ".join(cmd))
		try:
			check_call(cmd)
		except Exception as exc:
			print("Failed command", cmd, "with:", exc)
			if exists("install.log"):
				with open("install.log") as fh:
					print(fh.read())
		else:
			success = True
	if success:
		print("Installation complete!")
	else:
		print("Installation failed")


def download_python(version, arch):
	for _ in range(3):
		try:
			return download_file(URLS[version, arch], "installer.exe")
		except Exception as exc:
			print("Failed to download:", exc)
		print("Retrying ...")


def install_pip(home):
	pip_path = home + "/Scripts/pip.exe"
	python_path = home + "/python.exe"
	if exists(pip_path):
		print("pip already installed.")
	else:
		print("Installing pip...")
		download_file(GET_PIP_URL, GET_PIP_PATH)
		print("Executing:", python_path, GET_PIP_PATH)
		check_call([python_path, GET_PIP_PATH])


def install_packages(home, *packages):
	cmd = [home + "/Scripts/pip.exe", "install"]
	cmd.extend(packages)
	check_call(cmd)


if __name__ == "__main__":
	install_python(environ['PYTHON_VERSION'], environ['PYTHON_ARCH'], environ['PYTHON_HOME'])
	install_pip(environ['PYTHON_HOME'])
	install_packages(
		environ['PYTHON_HOME'],
		"setuptools>=18.0.1", "wheel", "tox", "virtualenv>=13.1.0", "nose", "rednose", "coverage", "codacy-coverage",
		"python-coveralls"
	)



from pyplanet.apps.config import AppConfig
from pyplanet.contrib.setting import Setting


class MigrationTest(AppConfig):
	SETTING_1 = Setting('test_setting_1', 'Setting 1', Setting.CAT_GENERAL)
	SETTING_2 = Setting('test_setting_2', 'Integer test', Setting.CAT_OTHER, type=int, default=-1)
	SETTING_3 = Setting('test_setting_3', 'Set test', Setting.CAT_FEATURES, type=set, default=set())
	SETTING_4 = Setting('test_setting_4', 'Enable this or that', Setting.CAT_FEATURES, type=bool, description='Large description that should break into multiple lines when displaying to the clients. Please test if this is the case at all.. Large description that should break into multiple lines when displaying to the clients. Please test if this is the case at all. Large description that should break into multiple lines when displaying to the clients. Please test if this is the case at all. Large description that should break into multiple lines when displaying to the clients. Please test if this is the case at all.')
	SETTING_5 = Setting('test_setting_5', 'Set test', Setting.CAT_FEATURES, type=set, default=set())

	async def on_start(self):
		await self.context.setting.register(
			self.SETTING_1, self.SETTING_2, self.SETTING_3, self.SETTING_4, self.SETTING_5
		)



"""
This folder contains the configuration in different files. This init file imports all files in the current directory
to make sure we combine configurations to one settings module.
Documentation: http://pypla.net/

If you want to use other configuration methods like YAML or JSON files, take a look at http://pypla.net/ and head to the
configuration pages.
"""

from .base import *
from .apps import *

try:
	from .local import *
except:
	pass



import asyncio
import logging
import uuid

from requests.exceptions import ConnectionError as RequestsConnectionError

from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.dedimania.api import DedimaniaAPI, DedimaniaRecord
from pyplanet.apps.contrib.dedimania.exceptions import DedimaniaException, DedimaniaTransportException, \
	DedimaniaNotSupportedException, DedimaniaFault, DedimaniaInvalidCredentials
from pyplanet.apps.core.trackmania import callbacks as tm_signals
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.apps.contrib.dedimania.api import DedimaniaAPI, DedimaniaRecord
from pyplanet.apps.contrib.dedimania.exceptions import (
	DedimaniaException, DedimaniaTransportException, DedimaniaNotSupportedException,
	DedimaniaFault, DedimaniaInvalidCredentials
)
from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting
from pyplanet.utils import times
from pyplanet.utils.log import handle_exception

from .views import DedimaniaRecordsWidget, DedimaniaRecordsListView

logger = logging.getLogger(__name__)


class Dedimania(AppConfig):
	game_dependencies = ['trackmania']
	app_dependencies = ['core.maniaplanet', 'core.trackmania']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.widget = None
		self.api = None

		self.lock = asyncio.Lock()
		self.current_records = []
		self.current_script = None
		self.player_info = dict()

		self.v_replay = None
		self.v_replay_checks = None
		self.ghost_replay = None
		self.ghost_files = []

		self.server_max_rank = None
		self.map_status = None
		self.map_uid = None
		self.ready = False

		self.setting_server_login = Setting(
			'server_login', 'Dedimania Server Login', Setting.CAT_KEYS, type=str,
			description='Only fill in when you want to override the auto-detected server login!',
			default=None, change_target=self.reload_settings
		)
		self.setting_dedimania_code = Setting(
			'dedimania_code', 'Dedimania Server Code', Setting.CAT_KEYS, type=str,
			description='The secret dedimania code. Get one at their site: $lhttp://dedimania.net/tm2stats/?do=register',
			default=None, change_target=self.reload_settings
		)

		self.setting_chat_welcome = Setting(
			'chat_welcome', 'Display welcome message', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether to display a welcome message indicating the ranking limits (server and player).',
			default=False
		)
		self.setting_chat_announce = Setting(
			'chat_announce', 'Minimum index for chat announce', Setting.CAT_BEHAVIOUR, type=int,
			description='Minimum record index needed for public new record/recordchange announcement (0 for disable).',
			default=50
		)
		self.setting_sent_announce = Setting(
			'sent_announce', 'Announce sending times to dedimania', Setting.CAT_BEHAVIOUR, type=bool,
			description='Enable the announce of successfully sent records to dedimania message.',
			default=False
		)

		self.login = self.code = self.server_version = self.pack_mask = None

	def is_mode_supported(self, mode):
		mode = mode.lower()
		return mode.startswith('timeattack') or mode.startswith('rounds') or mode.startswith('team') or \
			   mode.startswith('laps') or mode.startswith('cup')

	async def on_start(self):
		# Init settings.
		await self.context.setting.register(
			self.setting_server_login, self.setting_dedimania_code,
			self.setting_chat_welcome, self.setting_chat_announce, self.setting_sent_announce
		)

		# Load initial data.
		if not self.widget:
			self.widget = DedimaniaRecordsWidget(self)

		# Load settings + initiate api.
		if self.instance.game.game == "tm":
			await self.reload_settings()

		# Register signals
		self.context.signals.listen(mp_signals.map.map_begin, self.map_begin)
		self.context.signals.listen(mp_signals.map.map_start, self.map_start)
		self.context.signals.listen(mp_signals.map.map_end, self.map_end)

		self.context.signals.listen(mp_signals.flow.podium_start, self.podium_start)

		self.context.signals.listen(tm_signals.finish, self.player_finish)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.player.player_disconnect, self.player_disconnect)

		# Register commands
		await self.instance.command_manager.register(
			Command(
				'dedicps', target=self.command_dedicps, description='Compare your dedimania record checkpoints with another record.'
			).add_param('record', required=False, type=int, help='Custom record rank to compare with. Defaults to 1.', default=1)
		)

	async def reload_settings(self, *args, **kwargs):
		# Check setting + return errors if not correct!
		self.login = await self.setting_server_login.get_value(refresh=True) or self.instance.game.server_player_login
		self.code = await self.setting_dedimania_code.get_value(refresh=True)
		if not self.code:
			message = '$0b3Error: No dedimania code was provided, please edit the settings (//settings).'
			logger.error('Dedimania Code not configured! Please configure with //settings!')
			await self.instance.chat(message)
			return

		# Save current script name
		self.current_script = await self.instance.mode_manager.get_current_script()

		# Init API (execute this in a non waiting future).
		self.api = DedimaniaAPI(
			self.instance,
			self.login, self.code, self.instance.game.server_path, self.instance.map_manager.current_map.environment,
			self.instance.game.dedicated_version, self.instance.game.dedicated_build
		)
		asyncio.ensure_future(self.initiate_api())

	async def initiate_api(self):
		if not self.api:
			return
		await self.api.on_start()
		try:
			await self.api.authenticate()
			self.ready = True
		except RequestsConnectionError as e:
			logger.error('Can\'t connect to dedimania! Dedimania down or blocked by your host? {}'.format(str(e)))
			self.ready = False
			return
		except ConnectionRefusedError as e:
			logger.error('Can\'t connect to dedimania! Dedimania down or blocked by your host? {}'.format(str(e)))
			self.ready = False
			return
		except DedimaniaInvalidCredentials:
			logger.error('Can\'t connect to dedimania! Code or login is invalid!')
			message = '$f00Error: Your dedimania code or login is invalid!'
			await self.instance.chat(message)
			self.ready = False
			return
		except Exception as e:
			logger.exception(e)
			logger.error('Dedimania app unloaded!')
			return

		if self.ready:
			await self.map_begin(self.instance.map_manager.current_map)

	async def show_records_list(self, player, data = None, **kwargs):
		"""
		Show record list view to player.

		:param player: Player instance.
		:param data: -
		:param kwargs: -
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: view instance or nothing when there are no records.
		"""
		if not len(self.current_records):
			message = '$i$f00There are currently no dedimania records on this map!'
			await self.instance.chat(message, player)
			return

		view = DedimaniaRecordsListView(self)
		await view.display(player=player.login)
		return view

	async def map_start(self, map, restarted, **kwargs):
		if restarted:
			# Disable this line as the replays are already removed from the memory of the dedicated.
			# The method calling //restart should call this one from there.
			# await self.podium_start()

			# Clear the current map.
			self.map_uid = None

			await self.map_end(map)
			await self.map_begin(map)

	async def map_begin(self, map, **kwargs):
		# Reset.
		if not self.api:
			await self.reload_settings()
			await self.initiate_api()
		if not self.api:
			return

		self.api.retries = 0

		# Current script refresh. Save to local later because we need to compare it first.
		current_script = (await self.instance.mode_manager.get_current_script()).lower()

		# If the map uid already has been filled and the same we are starting double. Return immediately.
		# This is because of issue #276. Also check if the current script isn't changed, resolves #538.
		if self.map_uid == self.instance.map_manager.current_map.uid and current_script == self.current_script:
			return
		self.map_uid = self.instance.map_manager.current_map.uid

		# Set map status.
		self.map_status = map.author_login == 'nadeo' or map.time_author > 6200 and map.num_checkpoints > 1
		if not self.map_status:
			message = '$f90This map is not supported by Dedimania (min 1 checkpoint + 6.2 seconds or higher author time).'
			await self.widget.hide()
			return await self.instance.chat(message)

		# Refresh script.
		self.current_script = current_script

		# Fetch records + update widget.
		async with self.lock:
			self.v_replay = None
			self.v_replay_checks = None
			self.ghost_replay = None
			self.current_records = list()
		if self.ready:
			await self.widget.display()

		await self.refresh_records()

		if self.ready:
			await asyncio.gather(
				self.chat_current_record(),
				self.widget.display()
			)

		# Cleanup ghosts from previous maps.
		await self.cleanup()

		# Test map size.
		try:
			map_stat = await self.instance.storage.driver.stat('{}/{}'.format(
				self.instance.storage.MAP_FOLDER,
				map.file
			))
			if map_stat and map_stat.st_size and map_stat.st_size > 2750000:  # > 2.75Mb
				await self.instance.chat(
					'$f90Dedimania: The current map is large in size and might introduce problems with submitting records'
				)
		except:
			pass


	async def podium_start(self, force=False, **kwargs):
		# Get replays of the players.
		self.v_replay = None
		self.v_replay_checks = None

		self.ghost_replay = None

		async with self.lock:
			for pos, record in enumerate(self.current_records):
				if record.updated:
					if not self.v_replay:
						replay = await self.get_v_replay(record.login)
						if replay:
							self.v_replay = replay
					if not self.v_replay_checks and self.current_script.lower().startswith('laps'):
						self.v_replay_checks = ','.join([str(c) for c in record.race_cps])

					if pos == 0:
						replay = await self.get_ghost_replay(record.login)
						if replay:
							self.ghost_replay = replay

	async def map_end(self, map):
		if self.map_status is False:
			logger.warning('Don\'t send dedi records, map not supported!')
			return

		if not self.v_replay:
			return

		async with self.lock:
			try:
				logger.debug('Sending Dedimania times...')
				await self.api.set_map_times(
					map, self.current_script, self.current_records.copy(), self.v_replay, self.v_replay_checks, self.ghost_replay,
				)
				if await self.setting_sent_announce.get_value():
					await self.instance.chat(
						'$0b3Dedimania records have been sent successfully!'
					)
			except DedimaniaNotSupportedException:
				pass
			except (DedimaniaTransportException, DedimaniaFault) as e:
				logger.error(e)
				if 'session' not in str(e):
					await self.instance.gbx.multicall(
						self.instance.chat(
							'$f00Error: Dedimania got an error, didn\'t send records. '
							'There might be an issue with the map or the dedimania server!'
						),
						self.instance.chat(
							'$f00Error Details: {}'.format(str(e))
						)
					)
				else:
					await self.instance.chat(
						'$f00Error: Dedimania got an error, didn\'t send records. '
						'There might be an issue with the map or the dedimania server!'
					)

	async def cleanup(self):
		"""
		Cleanup old Ghost Replays.
		"""
		if not len(self.ghost_files):
			return

		coros = []
		for file in self.ghost_files:
			logger.debug('Removing old ghost file: {}'.format(file))
			coros.append(self.instance.storage.driver.remove(file))

		try:
			await asyncio.gather(*coros)
		except Exception as e:
			logger.warning('Errors with cleaning up dedimania ghost files (ignored):')
			logger.exception(e)

		self.ghost_files = list()

	async def refresh_records(self):
		try:
			player_list, self.current_script = await asyncio.gather(
				self.instance.gbx('GetPlayerList', -1, 0),
				self.instance.mode_manager.get_current_script(),
			)
			self.server_max_rank, modes, player_infos, self.current_records = await self.api.get_map_details(
				self.instance.map_manager.current_map,
				self.current_script,
				server_name=self.instance.game.server_name, server_comment='', is_private=self.instance.game.server_is_private,
				max_players=self.instance.game.server_max_players, max_specs=self.instance.game.server_max_specs,
				players=player_list,
				server_login=self.instance.game.server_player_login
			)
			self.ready = True
		except DedimaniaNotSupportedException as e:
			self.ready = False
			await self.instance.chat('$0b3Dedimania doesn\'t support or know the current script mode {}'.format(
				self.current_script
			))
			logger.warning('Dedimania doesn\'t support or known the mode {}'.format(self.current_script))

			# Still silently report.
			handle_exception(e, module_name=__name__, func_name='refresh_records', extra_data={
				'script': self.current_script
			})
			return
		except DedimaniaTransportException as e:
			self.ready = False

			if 'Max retries exceeded' in str(e):
				message = '$f00Error: Dedimania seems down?'
			else:
				message = '$f00Error: Dedimania error occured!'
				logger.exception(e)
			await self.instance.chat(message)
			return
		except DedimaniaFault as e:
			self.ready = False

			if 'session' in str(e).lower():
				handle_exception(e, module_name=__name__, func_name='refresh_records')
			logger.error('Dedimania gave an Fault: {}'.format(str(e)))

			self.current_records = list()
			return
		except Exception as e:
			self.ready = False
			handle_exception(e, module_name=__name__, func_name='refresh_records')
			logger.exception(e)
			self.current_records = list()
			return

		for info in player_infos:
			self.player_info[info['Login']] = dict(
				banned=False, login=info['Login'], max_rank=info['MaxRank'],
			)

	async def player_connect(self, player, is_spectator, **kwargs):
		try:
			if not self.api:
				return
			if self.ready:
				await self.widget.display(player=player)
			res = await self.instance.gbx('GetDetailedPlayerInfo', player.login)
			p_info = await self.api.player_connect(
				player.login, player.nickname, res['Path'], is_spectator
			)
			if p_info:
				self.player_info[player.login] = p_info

				if await self.setting_chat_welcome.get_value():
					message = '$0b3This server is using $l[http://dedimania.net/tm2stats/]Dedimania$l, current rank limits: $fff{}$0b3 (server), $fff{}$0b3 (player).'.format(
						int(self.server_max_rank), p_info['max_rank']
					)
					await self.instance.chat(message, player)

		except ConnectionRefusedError:
			return
		except TimeoutError:
			return
		except DedimaniaException:
			return

	async def player_disconnect(self, player, **kwargs):
		try:
			del self.player_info[player.login]
		except:
			pass
		try:
			if not self.api:
				return
			await self.api.player_disconnect(player.login, '')
		except:
			pass

	async def player_finish(self, player, race_time, lap_time, lap_cps, race_cps, flow, raw, **kwargs):
		if not self.map_status:
			return
		if not self.ready:
			return
		if player.login not in self.player_info:
			logger.warning('Player info not (yet) retrieved from dedimania for the player {}'.format(player.login))
			return
		player_info = self.player_info[player.login]
		chat_announce = await self.setting_chat_announce.get_value()
		current_records = [x for x in self.current_records if x.login == player.login]
		score = lap_time
		if len(current_records) > 0:
			current_record = current_records[0]
			previous_index = self.current_records.index(current_record) + 1

			if score < current_record.score:
				previous_time = current_record.score
				current_record.updated = True
				if self.current_script.lower().startswith('laps'):
					current_record.race_cps = race_cps
					current_record.cps = lap_cps

				# Determinate new rank.
				async with self.lock:
					new_rank = 1
					for idx, record in enumerate(self.current_records):
						new_rank = idx + 1
						if score <= record.score:
							break

				# Check if the player or server allows this player to finish with his max_rank.
				if new_rank > int(self.server_max_rank) and new_rank > int(player_info['max_rank']):
					logger.debug('Ignore time, not within server or player max_rank.')
					return

				# Update score + infos.
				async with self.lock:
					current_record.score = score
					current_record.cps = lap_cps
					self.current_records.sort(key=lambda x: x.score)

				if new_rank < previous_index:
					message = '$fff{}$z$s$0b3 gained the $fff{}.$0b3 Dedimania Record: $fff\uf017 {}$0b3 ($fff{}.$0b3 $fff-{}$0b3).'.format(
						player.nickname, new_rank, times.format_time(score), previous_index,
						times.format_time((previous_time - score))
					)
				else:
					message = '$fff{}$z$s$0b3 improved the $fff{}.$0b3 Dedimania Record: $fff\uf017 {}$0b3 ($fff-{}$0b3).'.format(
						player.nickname, new_rank, times.format_time(score),
						times.format_time((previous_time - score))
					)

				coros = [self.widget.display()]

				if chat_announce >= new_rank:
					coros.append(self.instance.chat(message))
				elif chat_announce != 0:
					coros.append(self.instance.chat(message, player))
				await asyncio.gather(*coros)

			elif score == current_record.score:
				message = '$fff{}$z$s$0b3 equalled the $fff{}.$0b3 Dedimania Record: $fff\uf017 {}$0b3.'.format(
					player.nickname, previous_index, times.format_time(score)
				)

				if chat_announce >= previous_index:
					return await self.instance.chat(message)
				elif chat_announce != 0:
					return await self.instance.chat(message, player)

		else:
			new_record = DedimaniaRecord(
				login=player.login, nickname=player.nickname, score=score, rank=None, max_rank=player_info['max_rank'],
				cps=lap_cps, vote=-1
			)
			new_record.updated = True
			if self.current_script.startswith('Laps'):
				new_record.race_cps = race_cps

			# Determinate new rank.
			async with self.lock:
				new_rank = 1
				for idx, record in enumerate(self.current_records):
					new_rank = idx + 1
					if score <= record.score:
						break

			# Check if the player or server allows this player to finish with his max_rank.
			if new_rank > int(self.server_max_rank) and new_rank > int(player_info['max_rank']):
				logger.debug('Ignore time, not within server or player max_rank.')
				return

			async with self.lock:
				self.current_records.append(new_record)
				self.current_records.sort(key=lambda x: x.score)
				new_index = self.current_records.index(new_record) + 1
			message = '$fff{}$z$s$0b3 drove the $fff{}.$0b3 Dedimania Record: $fff\uf017 {}$0b3.'.format(
				player.nickname, new_index, times.format_time(score)
			)

			if chat_announce >= new_index:
				chat_await = self.instance.chat(message)
			else:
				chat_await = self.instance.chat(message, player)

			await asyncio.gather(
				chat_await,
				self.widget.display(),
			)

	async def get_v_replay(self, login):
		try:
			return await self.instance.gbx('GetValidationReplay', login)
		except:
			return None

	async def get_ghost_replay(self, login):
		replay_name = 'dedimania_{}.Replay.Gbx'.format(uuid.uuid4().hex)
		try:
			await self.instance.gbx('SaveBestGhostsReplay', login, replay_name)
		except:
			return None
		try:
			async with self.instance.storage.open('UserData/Replays/{}'.format(replay_name)) as ghost_file:
				self.ghost_files.append('UserData/Replays/{}'.format(replay_name))
				return await ghost_file.read()
		except FileNotFoundError as e:
			message = '$f00Error: Dedimania requires you to have file access on the server. We can\'t fetch' \
					  'the driven replay!'
			logger.error('Please make sure we can access the dedicated files. Configure your storage driver correctly! '
						 '{}'.format(str(e)))
			await self.instance.chat(message)
			raise DedimaniaException('Can\'t access replay file')
		except PermissionError as e:
			message = '$f00Error: Dedimania requires you to have file access on the server. We can\'t fetch' \
					  'the driven replay because of an permission problem!'
			logger.error('We can\'t read files in the dedicated folder, your permissions don\'t allow us to read it! '
						 '{}'.format(str(e)))
			await self.instance.chat(message)
			raise DedimaniaException('Can\'t access files due to permission problems')

	async def chat_current_record(self):
		records_amount = len(self.current_records)
		if records_amount > 0:
			first_record = self.current_records[0]
			message = '$0b3Current Dedimania Record: $fff\uf017 {}$z$s$0b3 by $fff{}$z$s$0b3 ($l[http://dedimania.net/tm2stats/?do=stat&UId={}&Show=RECORDS]$fff{}$0b3 records$l)'.format(
				times.format_time(first_record.score), first_record.nickname, self.instance.map_manager.current_map.uid, records_amount
			)
			calls = list()
			calls.append(self.instance.chat(message))

			for player in self.instance.player_manager.online:
				calls.append(await self.chat_personal_record(player))
			return await asyncio.gather(*calls)
		else:
			message = '$0b3There is no Dedimania Record on this map yet.'
			return await self.instance.chat(message)

	async def chat_personal_record(self, player):
		async with self.lock:
			record = [x for x in self.current_records if x.login == player.login]

		if len(record) > 0:
			message = '$0b3You currently hold the $fff{}.$0b3 Dedimania Record: $fff\uf017 {}'.format(
				self.current_records.index(record[0]) + 1, times.format_time(record[0].score)
			)
			return self.instance.chat(message, player)
		else:
			message = '$0b3You don\'t have a Dedimania Record on this map yet.'
			return self.instance.chat(message, player)

	async def command_dedicps(self, player, data, *args, **kwargs):
		"""
		Compare dedi checkpoints from your record to another record, defaults to 1st record.

		:param player: Player instance
		:param data: Arguments data
		:param args: *
		:param kwargs: **
		:return:
		"""
		async with self.lock:
			record = [x for x in self.current_records if x.login == player.login]

			if data.record > len(self.current_records):
				message = '$0b3There is no record for rank {}!'.format(data.record)
				return await self.instance.chat(message, player)

			compare_record = self.current_records[data.record - 1]

		view = views.DedimaniaCpCompareListView(self, record[0] if len(record) else None, compare_record)
		await view.display(player)



import logging

from pyplanet import __version__ as version
from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.views.generics.widget import WidgetView
from pyplanet.utils import times
from pyplanet.contrib.player.exceptions import PlayerNotFound

logger = logging.getLogger(__name__)


class MapInfoWidget(WidgetView):
	widget_x = 125
	widget_y = 90
	z_index = 160

	template_name = 'info/mapinfo.xml'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widgets_mapinfo'

		self.mx_link_cache = dict()

	async def get_context_data(self):
		map = self.app.instance.map_manager.current_map

		# Load related info from other apps if installed and enabled, such as MX link.
		mx_link = None
		if 'mx' in self.app.instance.apps.apps:
			if map.uid in self.mx_link_cache:
				mx_link = self.mx_link_cache[map.uid]
			else:
				# Fetch, validate and make link.
				try:
					mx_info = await self.app.instance.apps.apps['mx'].api.map_info(map.uid)
					if mx_info and len(mx_info) >= 1:
						base_url = self.app.instance.apps.apps['mx'].api.base_url()
						mx_link = '{}/s/tr/{}'.format(
							base_url, mx_info[0][0]
						)
					self.mx_link_cache[map.uid] = mx_link
				except Exception as e:
					logger.error('Could not retrieve map info from MX/TM API for the info widget: {}'.format(str(e)))
					pass

		context = await super().get_context_data()
		context.update({
			'map_name': map.name,
			'map_mx_link': mx_link,
			'map_author': map.author_login,
			'map_authortime': times.format_time(map.time_author) if map.time_author and map.time_author > 0 else '-',
			'map_environment': map.environment,
		})

		return context


class ServerInfoWidget(WidgetView):
	widget_x = -160
	widget_y = 90
	z_index = 160

	template_name = 'info/serverinfo.xml'

	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.info.Info
		"""
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widgets_serverinfo'

	async def get_context_data(self):
		context = await super().get_context_data()

		ladder_min = int(self.app.instance.game.ladder_min)
		ladder_max = int(self.app.instance.game.ladder_max)

		if ladder_min > 1000:
			ladder_min = int(ladder_min / 1000)
		if ladder_max > 1000:
			ladder_max = int(ladder_max / 1000)

		context.update({
			'version': version,
			'num_players': self.app.instance.player_manager.count_players,
			'max_players': self.app.instance.player_manager.max_players,
			'num_spectators': self.app.instance.player_manager.count_spectators,
			'max_spectators': self.app.instance.player_manager.max_spectators,
			'ladder_min': ladder_min,
			'ladder_max': ladder_max,
		})

		return context



from pyplanet.utils import style


def test_style_stripping():
	# Colors
	raw = '$i$fffMax$06fSmurf$f00.$fffes$$l$09f.$fffm$08f$a5x$n$w$o'
	expect = '$iMaxSmurf.es$$l.m$a5x$n$w$o'
	assert style.style_strip(raw, style.STRIP_COLORS) == expect

	raw = '$l[some link]$i$FFFMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx$l'
	expect = '$l[some link]$iMaxSmurf.esl.mx$l'
	assert style.style_strip(raw, style.STRIP_COLORS) == expect

	raw = '$l[some link]$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx'
	expect = '$l[some link]$iMaxSmurf.esl.mx'
	assert style.style_strip(raw, style.STRIP_COLORS) == expect


	# Links
	raw = '$l$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x$l'
	expect = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect

	raw = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x'
	expect = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect

	raw = '$l[some link]$i$fffMax$06fSmurf$f00.$fffes$$l$09f.$fffm$08fx$l'
	expect = '$i$fffMax$06fSmurf$f00.$fffes$$09f.$fffm$08fx'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect

	raw = '$l[some link]$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx'
	expect = '$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08fx'
	assert style.style_strip(raw, style.STRIP_LINKS) == expect


	# Sizes
	raw = '$i$n$fffMax$06fSmurf$f00.$w$o$fffe$$nsl$09f.$w$fffm$08f$a5$ox'
	expect = '$fffMax$06fSmurf$f00.$fffe$sl$09f.$fffm$08f$a5x'
	assert style.style_strip(raw, style.STRIP_SIZES) == expect


	# All
	raw = '$h$i$fffMax$06fSmurf$f00.$fffesl$09f.$fffm$08f$a5x$h'
	expect = 'MaxSmurf.esl.m$a5x'
	assert style.style_strip(raw) == expect

	raw = '$l[some link]$i$fffMax$06fSmur$$f$f00.$fffesl$09f.$fffm$08fx$l'
	expect = 'MaxSmur$$f.esl.mx'
	assert style.style_strip(raw) == expect

	raw = '$l[some link]$i$fffMax$06fSmu$nrf$f00.$fffesl$09f.$fffm$08fx'
	expect = 'MaxSmurf.esl.mx'
	assert style.style_strip(raw) == expect


	# Combination & Misc
	raw = '$f80$i$oToffe$g$z$06fSmurf $z$sHello'
	expect = 'Toffe$zSmurf $zHello'
	assert style.style_strip(raw, keep_reset=True) == expect

	raw = '$f80$i$oToffe$g$z$06fSmurf $z$sHello'
	expect = 'ToffeSmurf $sHello'
	assert style.style_strip(raw, style.STRIP_SIZES, style.STRIP_COLORS) == expect



from pyplanet.views.generics.widget import WidgetView


class VoteWidget(WidgetView):
	"""
	Voting widget.
	"""

	widget_x = -25
	widget_y = -58.5
	template_name = 'voting/voting.xml'

	def __init__(self, app):
		"""
		Initializes the widget.

		:param app: the voting application (plugin)
		"""
		
		super().__init__(app.context.ui)
		self.app = app
		self.id = 'pyplanet__widgets_voting'

		self.subscribe('yes', self.vote_yes)
		self.subscribe('no', self.vote_no)

	async def get_context_data(self):
		"""
		Called to request data to display in the widget.
		Will return the type of vote currently running.
		"""

		context = await super().get_context_data()

		context.update({
			'vote_type': self.app.current_vote.action
		})

		return context

	async def vote_yes(self, player, *args, **kwargs):
		"""
		Called when a player votes yes via the widget (either via button or F5).

		:param player: player who voted
		"""

		await self.app.vote_yes(player, None)

	async def vote_no(self, player, *args, **kwargs):
		"""
		Called when a player votes no via the widget (either via button or F6).

		:param player: player who voted
		"""
		
		await self.app.vote_no(player, None)



import threading
import logging

from collections import OrderedDict

from pyplanet.utils.toposort import toposort
from pyplanet.apps.config import AppConfig, AppState
from pyplanet.core.exceptions import ImproperlyConfigured


class Apps:
	"""
	The apps class contains the context applications, loaded or not loaded in order of declaration or requirements
	if given by app configuration.

	The apps should contain a configuration class that could be loaded for reading out metadata, options and other
	useful information such as description, author, version and more.
	"""

	def __init__(self, instance):
		"""
		Initiate registry with pre-loaded apps.

		:param instance: Instance of the controller.
		:type instance: pyplanet.core.instance.Instance
		"""
		self.instance = instance

		self.apps = OrderedDict()
		self.unloaded_apps = OrderedDict()

		# Set ready states.
		self.apps_ready = self.ready = False

		# Set a lock for threading.
		self._lock = threading.Lock()

		# Listen to events
		self.instance.signals.listen('contrib.mode:script_mode_changed', self._on_mode_change)

	def populate(self, apps, in_order=False):
		"""
		Loads application into the apps registry. Once you populate, the order isn't yet been decided.
		After all imports are done you should shuffle the apps list so it's in the right order of execution!

		:param apps: Apps list.
		:param in_order: Is the list already in order?
		:type apps: list
		"""
		if self.ready:
			return

		populated_apps = dict()
		dep_dict = dict()

		# Load modules.
		for entry in apps:
			app = AppConfig.import_app(entry, self.instance)

			# Check if the app is unique.
			if app.label in self.apps:
				raise ImproperlyConfigured('Application labels aren\'t unique! Duplicates: {}'.format(app.label))

			# Inject apps instance into app itself.
			app.apps = self

			# Set state on app.
			app.state = AppState.UNLOADED

			# Get dependencies to other apps.
			deps = getattr(app, 'app_dependencies', list())
			if not type(deps) is list:
				deps = list()

			# Add to the list so it can get ordered by dependencies. (not if in_order is true).
			if in_order:
				self.apps[app.label] = app
			else:
				populated_apps[app.label] = app

				# Add nodes of dependencies.
				dep_dict[app.label] = deps

		if in_order:
			return

		# Determinate order.
		order = toposort(dep_dict)

		# Add in order
		for label in order:
			try:
				self.apps[label] = populated_apps[label]
			except KeyError:
				if label.startswith('core.'):
					pass
				else:
					raise Exception('One of the apps depends on a non existing app: {}'.format(label))

	async def check(self, on_start=False):
		"""
		Check and remove unsupported apps based on the current game and script mode. Also loads unloaded apps and try
		if the mode and game does support it again.
		"""
		# Check if disabled apps can be loaded again.
		for app_label, app_module in list(self.unloaded_apps.items()):
			try:
				# Load the module and initiate by creating the app class instance.
				self.populate([app_module], in_order=True)
				if app_label not in self.apps:
					raise Exception()  # Flow control, stop executing restart of app.

				# Init + start the app again.
				await self.apps[app_label].on_init()
				await self.apps[app_label].on_start()

				# Clear the label from the unloaded list.
				del self.unloaded_apps[app_label]

				logging.info('(Re)loaded app {} as it seems that it supports this game/mode again.'.format(app_label))
			except Exception as e:
				logging.debug('Can\'t start app {}, Got exception with error: {}'.format(app_label, str(e)))
				# logging.exception(e)
				# Some apps can't be reloaded.
				pass

		# Check enabled apps, and replace the apps dictionary with the up-to-date apps.
		script_name = await self.instance.mode_manager.get_current_script(refresh=True)
		apps_dict = OrderedDict()
		for label, app in self.apps.items():
			if not app.is_game_supported(self.instance.game.game_full):
				logging.warning('Unloading app {}. Doesn\'t support the current game \'{}\'!'.format(
					label, self.instance.game.game_full
				))
				if not on_start:
					await app.on_stop()
				await app.on_destroy()

				self.unloaded_apps[label] = app.module.__name__
				del app

			elif not app.is_mode_supported(script_name):
				logging.warning('Unloading app {}. Doesn\'t support the current script mode!'.format(label))
				if not on_start:
					await app.on_stop()
				await app.on_destroy()

				self.unloaded_apps[label] = app.module.__name__
				del app

			else:
				apps_dict[label] = app

		self.apps = apps_dict

	async def discover(self):
		"""
		The discover function will discover all models, signals and more
		from apps in the right order.
		"""
		for label, app in self.apps.items():
			# Discover models.
			self.instance.db.registry.init_app(app)

			# Discover signals.
			self.instance.signals.init_app(app)

		# Finishing signal manager.
		self.instance.signals.finish_reservations()

	async def init(self):
		"""
		This method will initiate all apps in order and in series.
		"""
		if self.apps_ready:
			raise Exception('Apps are not yet ordered!')
		for label, app in self.apps.items():
			await app.on_init()

	async def start(self):
		"""
		This method will start all apps that are previously initiated.
		"""
		if self.apps_ready:
			raise Exception('Apps are not yet ordered!')

		# The apps are in order, lets loop over them.
		for label, app in self.apps.items():
			await app.on_start()
			app.state = AppState.LOADED
			logging.debug('App is ready: {}'.format(label))
		logging.info('Apps successfully started!')

	async def stop(self):
		"""
		This method is executed when the instance is shutting down (will stop all the apps).
		"""
		for label, app in reversed(self.apps.items()):
			if app.state == AppState.LOADED:
				await app.on_stop()
				logging.debug('Stopped app {}'.format(label))
		logging.info('Apps successfully stopped!')

	async def _on_mode_change(self, unloaded_script, loaded_script, **kwargs):
		await self.check()



import asyncio

from pyplanet.apps.core.maniaplanet.callbacks.other import server_chat
from pyplanet.core.events import Callback, Signal
from pyplanet.core.exceptions import SignalGlueStop
from pyplanet.core.instance import Controller


async def handle_player_connect(source, signal, **kwargs):
	player_login, is_spectator = source
	player = await Controller.instance.player_manager.handle_connect(login=player_login)
	if not player:
		raise SignalGlueStop()
	return dict(
		player=player, is_spectator=is_spectator, source=source, signal=signal,
	)

async def handle_player_disconnect(source, signal, **kwargs):
	player_login, reason = source
	player = await Controller.instance.player_manager.handle_disconnect(login=player_login)
	if not player:
		raise SignalGlueStop()
	return dict(
		player=player, reason=reason, source=source, signal=signal,
	)

async def handle_player_chat(source, signal, **kwargs):
	player_uid, player_login, text, cmd = source
	if Controller.instance.game.server_player_login == player_login and Controller.instance.game.server_is_dedicated:
		# Inform our server_chat signal.
		asyncio.ensure_future(server_chat.send_robust(dict(text=text, cmd=cmd)))
		raise SignalGlueStop('We won\'t inform anything about the chat we send ourself!')
	try:
		player = await Controller.instance.player_manager.get_player(login=player_login, lock=True)
	except:
		raise SignalGlueStop()
	return dict(
		player=player, text=text, cmd=cmd
	)


async def handle_player_info_changed(source, signal, **kwargs):
	# Unpack spectator status.
	is_spectator =     		bool(source['SpectatorStatus']			% 10)
	is_temp_spectator =		bool((source['SpectatorStatus'] / 10)	% 10)
	is_pure_spectator =		bool((source['SpectatorStatus'] / 100)	% 10)
	auto_target =			bool((source['SpectatorStatus'] / 1000)	% 10)
	target_id =				int((source['SpectatorStatus'] // 10000))

	# Unpack flags.
	force_spectator =				int((source['Flags'] % 10))  # Int
	is_referee =					bool((source['Flags'] / 10)			% 10)
	is_podium_ready =				bool((source['Flags'] / 100)		% 10)
	is_using_stereoscopy =			bool((source['Flags'] / 1000)		% 10)
	is_managed_by_other_server =	bool((source['Flags'] / 10000)		% 10)
	is_server =						bool((source['Flags'] / 100000)		% 10)
	has_player_slot =				bool((source['Flags'] / 1000000)	% 10)
	is_broadcasting =				bool((source['Flags'] / 10000000)	% 10)
	has_joined_game =				bool((source['Flags'] / 100000000)	% 10)

	# https://github.com/maniaplanet/dedicated-server-api/blob/master/libraries/Maniaplanet/DedicatedServer/Structures/PlayerInfo.php#L69

	try:
		player = await Controller.instance.player_manager.get_player(login=source['Login'])
	except:
		player = None
	try:
		target = await Controller.instance.player_manager.get_player_by_id(target_id)
	except:
		target = None

	payload = dict(
		is_spectator=is_spectator, is_temp_spectator=is_temp_spectator, is_pure_spectator=is_pure_spectator,
		auto_target=auto_target, target_id=target_id, target=target, flags=source['Flags'],
		spectator_status=source['SpectatorStatus'], team_id=source['TeamId'],
		player_id=source['PlayerId'], player=player, player_login=source['Login'],

		# Since 0.6.0:
		force_spectator=force_spectator, is_referee=is_referee, is_podium_ready=is_podium_ready,
		is_using_stereoscopy=is_using_stereoscopy, is_managed_by_other_server=is_managed_by_other_server,
		is_server=is_server, has_player_slot=has_player_slot, is_broadcasting=is_broadcasting,
		has_joined_game=has_joined_game,
	)

	# Handle changes in our player_manager.
	await Controller.instance.player_manager.handle_info_change(**payload)

	return payload


player_connect = Callback(
	call='ManiaPlanet.PlayerConnect',
	namespace='maniaplanet',
	code='player_connect',
	target=handle_player_connect,
)
"""
:Signal:
	Player has been connected.
:Code:
	``maniaplanet:player_connect``
:Description:
	Callback sent when a player connects and we fetched our data.
:Original Callback:
	`Native` Maniaplanet.PlayerConnect

:param player: Player instance
:param is_spectator: Boolean determinating if the player joined as spectator.
:param source: Raw payload, best to not use!
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""

player_disconnect = Callback(
	call='ManiaPlanet.PlayerDisconnect',
	namespace='maniaplanet',
	code='player_disconnect',
	target=handle_player_disconnect,
)
"""
:Signal:
	Player has been disconnected.
:Code:
	``maniaplanet:player_disconnect``
:Description:
	Callback sent when a player disconnects.
:Original Callback:
	`Native` Maniaplanet.PlayerDisconnect

:param player: Player instance
:param reason: Reason of leave
:param source: Raw payload, best to not use!
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""

player_chat = Callback(
	call='ManiaPlanet.PlayerChat',
	namespace='maniaplanet',
	code='player_chat',
	target=handle_player_chat,
)
"""
:Signal:
	Player has been writing a chat entry. When the server writes something we **wont** inform it in here!
:Code:
	``maniaplanet:player_chat``
:Description:
	Callback sent when a player chats.
:Original Callback:
	`Native` Maniaplanet.PlayerChat

:param player: Player instance
:param text: Text of chat
:param cmd: Boolean if it's a command. Be aware, you should use the ``command`` manager for commands!
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_info_changed = Callback(
	call='ManiaPlanet.PlayerInfoChanged',
	namespace='maniaplanet',
	code='player_info_changed',
	target=handle_player_info_changed,
)
"""
:Signal:
	Player has changed status.
:Code:
	``maniaplanet:player_info_changed``
:Description:
	Callback sent when a player changes from state or information. The callback has been updated in 0.6.0 to include the
	information retrieved from extracting the flags parameter.
:Original Callback:
	`Native` Maniaplanet.PlayerInfoChanged

:param player: Player instance (COULD BE NONE SOMETIMES!).
:param player_login: Player login string.
:param is_spectator: Is player spectator (bool).
:param is_temp_spectator: Is player temporary spectator (bool).
:param is_pure_spectator: Is player pure spectator (bool).
:param auto_target: Player using auto target.
:param target_id: The target player id (not login!).
:param target: The target player instance or None if not found/none spectating.
:param flags: Raw flags.
:param spectator_status: Raw spectator status.
:param team_id: Team ID of player.
:param player_id: Player ID (server id).
:param force_spectator: 1, 2 or 3. Force spectator state
:param is_referee: Is the player a referee.
:param is_podium_ready: Is the player podium ready.
:param is_using_stereoscopy: Is the player using stereoscopy
:param is_managed_by_other_server: Is the player managed by another server (relaying).
:param is_server: Is the player one of the servers.
:param has_player_slot: Has the player a reserved player slot.
:param is_broadcasting: Is the player broadcasting (steaming) via the in-game stream functionality.
:param has_joined_game: Is the player ready and has it joined the game as player.
:type force_spectator: int
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
:type target: pyplanet.apps.core.maniaplanet.models.player.Player
"""

player_enter_player_slot = Signal(
	namespace='maniaplanet',
	code='player_enter_player_slot',
)
"""
:Signal:
	Player enters a player slot.
:Code:
	``maniaplanet:player_enter_player_slot``
:Description:
	Player change into a player, is using a player slot.
:Original Callback:
	*None*

:param player: Player instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""

player_enter_spectator_slot = Signal(
	namespace='maniaplanet',
	code='player_enter_spectator_slot',
)
"""
:Signal:
	Player enters a spectator slot (not temporary).
:Code:
	``maniaplanet:player_enter_spectator_slot``
:Description:
	Player change into a spectator, is using a spectator slot.
:Original Callback:
	*None*

:param player: Player instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""




class _Game:
	"""
	The game class holds information about the game itself and the server. The properties can be virtually overriden
	to be able to proxy to new/old syntaxes. This way we can provide a read-only data structure and still
	maintain the same structure if any of the third party API changes.

	This class is available from the instance with `instance.game`.

	**Most variables seem to contain None, but they actually get propagated during the start and connection
	with the dedicated server**
	"""
	dedicated_build = None
	dedicated_version = None
	dedicated_api_version = None
	dedicated_title = None

	server_is_dedicated = None
	server_is_server = None
	server_is_private = None
	server_ip = None
	server_p2p_port = None
	server_port = None
	server_player_login = None
	server_player_id = None
	server_download_rate = None
	server_upload_rate = None
	server_data_dir = None
	server_map_dir = None
	server_skin_dir = None
	server_language = None
	server_name = None
	server_path = None
	server_password = None
	server_spec_password = None
	server_max_players = None
	server_next_max_players = None
	server_max_specs = None
	server_next_max_specs = None

	ladder_min = None
	ladder_max = None

	game = None  # tm / sm / tmnext

	def game_from_environment(self, environment, game_name=None, title_id=None):
		if game_name == 'Trackmania' and title_id == 'Trackmania':
			return 'tmnext'
		if environment in ['Canyon', 'Stadium', 'Valley', 'Lagoon']:
			return 'tm'
		return 'sm'

	@property
	def game_full(self):
		if self.game == 'tm':
			return 'trackmania'
		elif self.game == 'tmnext':
			return 'trackmania_next'
		return 'shootmania'

Game = _Game()




from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.statistics.mp import ManiaplanetComponent
from pyplanet.apps.core.statistics.processor import StatisticsProcessor
from pyplanet.apps.core.statistics.tm import TrackmaniaComponent


class Statistics(AppConfig):
	core = True

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		# Initiate components.
		self.processor = StatisticsProcessor(self)
		self.trackmania = TrackmaniaComponent(self)
		self.maniaplanet = ManiaplanetComponent(self)

	async def on_init(self):
		# Call components.
		if self.instance.game.game in ['tm', 'tmnext']:
			await self.trackmania.on_init()

		await self.maniaplanet.on_init()

	async def on_start(self):
		# Call components.
		if self.instance.game.game in ['tm', 'tmnext']:
			await self.trackmania.on_start()

		await self.maniaplanet.on_start()



import asynctest

from pyplanet.core import Controller
from pyplanet.core.storage import StorageDriver


class TestStorageManager(asynctest.TestCase):
	async def test_init(self):
		instance = Controller.prepare(name='default').instance
		assert instance.storage
		assert instance.storage.driver
		assert isinstance(instance.storage.driver, StorageDriver)

	async def test_driver_interface(self):
		instance = Controller.prepare(name='default').instance
		assert type(instance.storage.driver.openable()) is bool



from pyplanet.views import TemplateView


class DiscordLogoView(TemplateView):
	widget_x = 0
	widget_y = 90
	size_x = 50
	size_y = 50
	template_name = 'ads/discord_logo.xml'

	def __init__(self, app, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.app = app
		self.id = 'discord__logo'

	async def get_context_data(self):
		context = await super().get_context_data()
		context.update({
			'discord_url': self.app.discord_join_url
		})
		return context

	async def display(self, **kwargs):
		return await super().display(**kwargs)


class PayPalLogoView(TemplateView):
	widget_x = 0
	widget_y = 0
	template_name = 'ads/paypal_logo.xml'

	def __init__(self, app, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.app = app
		self.id = 'paypal__logo'

	async def get_context_data(self):
		if await self.app.setting_enable_discord.get_value():
			self.widget_x = 131
			self.widget_y = -50
		else:
			self.widget_x = 139
			self.widget_y = -50
		context = await super().get_context_data()
		context.update({
			'donation_url': self.app.paypal_donation_url,
			'pos_x': self.widget_x,
			'pos_y': self.widget_y
		})
		return context

	async def display(self, **kwargs):
		return await super().display(**kwargs)



from pyplanet.core import Controller
from pyplanet.core.events import Callback, handle_generic, Signal
from pyplanet.contrib.player.exceptions import PlayerNotFound

async def handle_echo(source, signal, **kwargs):
	internal, public = source
	return dict(
		internal=internal, public=public
	)

async def handle_bill_updated(source, signal, **kwargs):
	bill_id, state, state_name, transaction_id = source
	return dict(
		bill_id=bill_id, state=state, state_name=state_name, transaction_id=transaction_id
	)

async def handle_vote_updated(source, signal, **kwargs):
	state, login, cmd_name, cmd_param = source
	player = None
	try:
		player = await Controller.instance.player_manager.get_player(login)
	except PlayerNotFound:
		player = None
	return dict(
		player=player, state=state, cmd_name=cmd_name, cmd_param=cmd_param
	)

bill_updated = Callback(
	call='ManiaPlanet.BillUpdated',
	namespace='maniaplanet',
	code='bill_updated',
	target=handle_bill_updated
)
"""
:Signal:
	Bill has been updated.
:Code:
	``maniaplanet:bill_updated``
:Description:
	Callback sent when a bill has been updated.
:Original Callback:
	`Native` Maniaplanet.BillUpdated

:param 1: Bill id.
:param 2: State.
:param 3: State name.
:param 4: Transaction id.
:type 1: int
:type 2: int
:type 3: str
:type 4: int
"""

vote_updated = Callback(
	call='ManiaPlanet.VoteUpdated',
	namespace='maniaplanet',
	code='vote_updated',
	target=handle_vote_updated
)
"""
:Signal:
	Vote has been updated.
:Code:
	``maniaplanet:vote_updated``
:Description:
	Callback sent when a call vote has been updated.
:Original Callback:
	`Native` Maniaplanet.VoteUpdated

:param player: Player instance
:param state: State name
:param cmd_name: Command name
:param cmd_param: Parameter given with command.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


server_chat = Signal(
	namespace='maniaplanet',
	code='server_chat'
)
"""
:Signal:
	Server send a chat message.
:Code:
	``maniaplanet:server_chat``
:Description:
	Custom signal called when the server outputs a message.
:Origin Callback:
	None (via Chat callback).
"""


server_password = Signal(
	namespace='maniaplanet',
	code='server_password'
)
"""
:Signal:
	Server player or spectator password changed
:Code:
	``maniaplanet:server_password``
:Description:
	Custom signal called when the password has been changed with PyPlanet.
:Origin Callback:
	None.

:param password: String with the new password.
:param kind: Kind of password, could be 'player' or 'spectator'.
:type password: str
:type kind: str
"""


channel_progression_start = Callback(
	call='Script.ManiaPlanet.ChannelProgression_Start',
	namespace='maniaplanet',
	code='channel_progression_start',
	target=handle_generic
)
"""
:Signal:
	Signal sent when channel progression sequence starts.
:Code:
	``maniaplanet:channel_progression_start``
:Description:
	Callback sent when the channel progression sequence starts.
:Original Callback:
	`Script` Maniaplanet.ChannelProgression_Start

:param time: Time when callback has been sent.
"""


channel_progression_end = Callback(
	call='Script.ManiaPlanet.ChannelProgression_End',
	namespace='maniaplanet',
	code='channel_progression_end',
	target=handle_generic
)
"""
:Signal:
	Signal sent when channel progression sequence ends.
:Code:
	``maniaplanet:channel_progression_end``
:Description:
	Callback sent when the channel progression sequence ends.
:Original Callback:
	`Script` Maniaplanet.ChannelProgression_End

:param time: Time when callback has been sent.
"""

on_echo = Callback(
	call='ManiaPlanet.Echo',
	namespace='maniaplanet',
	code='on_echo',
	target=handle_echo
)
"""
:Signal:
	Echo was sent from other Controller/GBXRemote.
:Code:
	``maniaplanet:on_echo``
:Description:
	Callback sent when a echo was sent from Controller/GBXRemote.
:Original Callback:
	`Native` Maniaplanet.Echo

Echo('Test1', 'Test2')
will be reverted to:
('Test2, Test1')

:param internal: internal
:param public: public
"""



import logging

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.trackmania import callbacks as tm_signals
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

from .view import CPWidgetView


class CurrentCPs(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next']
	app_dependencies = ['core.maniaplanet', 'core.trackmania']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.current_cps = {}  # Maps a player login to a PlayerCP object
		self.widget = None
		self.player_cps = []  # Holds the sorted PlayerCP objects (these get displayed by the widget)

		self.dedimania_enabled = False

	# FOR TESTING ONLY, DO NOT USE IN PRODUCTION CODE
	def fill_test_cps(self):
		"""
			Adds some virtual players to the current cps
		"""
		self.current_cps['test1'] = PlayerCP(VirtualPlayer('$b$wTesT1', 'test1'), 4, 74213)
		self.current_cps['test2'] = PlayerCP(VirtualPlayer('$b$wTesT2', 'test2'), -1, 50214)
		self.current_cps['test3'] = PlayerCP(VirtualPlayer('$b$wTesT3', 'test3'), 2, 74215)
		self.current_cps['test4'] = PlayerCP(VirtualPlayer('$b$wTesT4', 'test4'), -1, 62216)
		self.current_cps['test5'] = PlayerCP(VirtualPlayer('$b$wTesT5', 'test5'), 6, 74217)
		self.current_cps['test6'] = PlayerCP(VirtualPlayer('$b$wTesT6', 'test6'), -1, 54218)
		self.current_cps['test7'] = PlayerCP(VirtualPlayer('$b$wTesT7', 'test7'), 1, 74219)
		self.current_cps['test8'] = PlayerCP(VirtualPlayer('$b$wTesT8', 'test8'), 2, 74220)
		self.current_cps['test9'] = PlayerCP(VirtualPlayer('$b$wTesT9', 'test9'), 3, 74221)
		self.current_cps['test10'] = PlayerCP(VirtualPlayer('$b$wTesT10', 'test10'), 6, 74222)
		self.current_cps['test11'] = PlayerCP(VirtualPlayer('$b$wTesT11', 'test11'), 5, 74223)
		self.current_cps['test12'] = PlayerCP(VirtualPlayer('$b$wTesT12', 'test12'), -1, 60225)
		self.current_cps['test13'] = PlayerCP(VirtualPlayer('$b$wTesT13', 'test13'), -1, 64226)
		self.current_cps['test14'] = PlayerCP(VirtualPlayer('$b$wTesT14', 'test14'), 5, 74227)
		self.current_cps['test15'] = PlayerCP(VirtualPlayer('$b$wTesT15', 'test15'), -1, 70228)
		self.current_cps['test16'] = PlayerCP(VirtualPlayer('$b$wTesT16', 'test16'), 3, 74229)
		self.current_cps['test17'] = PlayerCP(VirtualPlayer('$b$wTesT17', 'test17'), -1, 54230)
		self.current_cps['test18'] = PlayerCP(VirtualPlayer('$b$wTesT18', 'test18'), 3, 74231)
		self.current_cps['test19'] = PlayerCP(VirtualPlayer('$b$wTesT19', 'test19'), 4, 74232)
		self.current_cps['test20'] = PlayerCP(VirtualPlayer('$b$wTesT20', 'test20'), 5, 74233)

	async def on_start(self):
		# Listen to some signals
		self.context.signals.listen(tm_signals.waypoint, self.player_cp)
		self.context.signals.listen(tm_signals.start_line, self.player_start)
		self.context.signals.listen(tm_signals.finish, self.player_finish)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.player.player_disconnect, self.player_disconnect)
		self.context.signals.listen(mp_signals.map.map_start__end, self.map_start)
		self.context.signals.listen(mp_signals.player.player_enter_spectator_slot, self.player_enter_spec)

		# Make sure we move the rounds_scores and other gui elements.
		if self.instance.game.game in ['tm', 'sm']:
			self.instance.ui_manager.properties.set_attribute('round_scores', 'pos', '-126.5 80. 150.')
			self.instance.ui_manager.properties.set_attribute('multilap_info', 'pos', '107., 88., 5.')

		self.dedimania_enabled = ('dedimania' in self.instance.apps.apps and 'dedimania' not in self.instance.apps.unloaded_apps)

		self.widget = CPWidgetView(self)
		# await self.widget.display()
		await self.update_view()

	# When a player passes a CP
	async def player_cp(self, player, race_time, raw, *args, **kwargs):
		cp = int(raw['checkpointinrace'])  # Have to use raw to get the current CP
		# Create new PlayerCP object if there is no PlayerCP object for that player yet
		if player.login not in self.current_cps:
			self.current_cps[player.login] = PlayerCP(player)
		self.current_cps[player.login].cp = cp + 1  # +1 because checkpointinrace starts at 0
		self.current_cps[player.login].time = race_time
		await self.update_view()

	# When a player starts the race
	async def player_start(self, player, *args, **kwargs):
		if player.login not in self.current_cps:
			self.current_cps[player.login] = PlayerCP(player)
		await self.update_view()

	# When a player passes the finish line
	async def player_finish(self, player, race_time, race_cps, is_end_race, raw, *args, **kwargs):
		# Create new PlayerCP object if there is no PlayerCP object for that player yet
		if player.login not in self.current_cps:
			self.current_cps[player.login] = PlayerCP(player)

		# Set the current CP to -1 (signals finished) when a player finishes the race
		if is_end_race:
			self.current_cps[player.login].cp = -1
		else:
			self.current_cps[player.login].cp = int(raw['checkpointinrace']) + 1  # Otherwise just update the current cp
			# logging.debug(raw)
		self.current_cps[player.login].time = race_time
		await self.update_view()

	# When a player connects
	async def player_connect(self, player, *args, **kwargs):
		await self.update_view()
		logging.debug("Player connected: " + player.login)

	# When a player disconnects
	async def player_disconnect(self, player, *args, **kwargs):
		# Remove the current CP from the widget when a player leaves the server
		self.current_cps.pop(player.login, None)
		await self.update_view()

	# When a player enters spectator mode
	async def player_enter_spec(self, player, *args, **kwargs):
		# Remove the current CP from the widget when a player starts to spectate
		self.current_cps.pop(player.login, None)
		await self.update_view()

	# When the map start (end of event)
	async def map_start(self, *args, **kwargs):
		self.dedimania_enabled = ('dedimania' in self.instance.apps.apps and 'dedimania' not in self.instance.apps.unloaded_apps)
		self.current_cps.clear()  # Clear the current CPs when the map ends
		await self.update_view()

	# Update the view for all players
	async def update_view(self):
		# Used for sorting the PlayerCP objects by the 1. CP and 2. the time (Finished players are always on top)
		def keyfunc(key):
			lpcp = self.current_cps[key]
			return 1 if lpcp.cp == -1 else 2, -lpcp.cp, lpcp.time

		self.player_cps.clear()

		# Sort the PlayerCP objects by using the key function above and copy them into the player_cps-list
		for login in sorted(self.current_cps, key=lambda x: keyfunc(x)):
			pcp = self.current_cps[login]
			cp = pcp.cp
			cpstr = str(cp)
			if cp == -1:
				cpstr = "fin"
			self.player_cps.append(pcp)

		await self.widget.display()  # Update the widget for all players

	async def spec_player(self, player, target_login):
		logging.debug(player.login + ' will now spec ' + target_login)
		logging.debug(await self.instance.gbx.multicall(
			self.instance.gbx('ForceSpectator', player.login, 3),
			self.instance.gbx('ForceSpectatorTarget', player.login, target_login, -1)
		))


class PlayerCP:
	def __init__(self, player, cp=0, time=0):
		self.player = player
		self.cp = cp
		self.time = time


class VirtualPlayer:
	def __init__(self, nickname, login):
		self.nickname = nickname
		self.login = login



import asyncio
import gc
import logging

from pyplanet.utils import log

logger = logging.getLogger(__name__)


class _LeakChecker:
	def __init__(self):
		self.reported = False

	def start(self):
		asyncio.get_event_loop().call_later(5, self.check_memory_leak)

	def check_memory_leak(self):
		asyncio.get_event_loop().call_later(30, self.check_memory_leak)
		logger.debug('Checking for memory leaks...')
		if 0 < len(gc.garbage) < 20:
			logger.warning('Found possible memory leaks: {}'.format(gc.garbage))
		elif len(gc.garbage) >= 20:
			logger.error(
				'Found memory leaks: {}'.format(gc.garbage)
			)
			if not self.reported:
				try:
					raise MemoryError('Found memory leaks: {}'.format(gc.garbage))
				except MemoryError as e:
					log.handle_exception(exception=e, extra_data=dict(leaks=gc.garbage))
				self.reported = True


checker = _LeakChecker()



import asyncio
import os
import logging
import re

from xmlrpc.client import Fault
from peewee import DoesNotExist

from pyplanet.utils.log import handle_exception
from pyplanet.apps.core.maniaplanet.models import Map
from pyplanet.conf import settings
from pyplanet.contrib import CoreContrib
from pyplanet.contrib.map.exceptions import MapNotFound, MapException, ModeIncompatible
from pyplanet.core.exceptions import ImproperlyConfigured


class MapManager(CoreContrib):
	"""
	Map Manager. Manages the current map pool and the current and next map.

	.. todo::

		Write introduction.

	.. warning::

		Don't initiate this class yourself.

	"""
	def __init__(self, instance):
		"""
		Initiate, should only be done from the core instance.

		:param instance: Instance.
		:type instance: pyplanet.core.instance.Instance
		"""
		self._instance = instance
		self.lock = asyncio.Lock()

		# The matchsettings contains the name of the current loaded matchsettings file.
		self._matchsettings = None

		# The maps contain a list of map instances in the order that are in the current loaded list.
		self._maps = set()

		# The current map will always be in this variable. The next map will always be here. It will be updated. once
		# it's updated it should be send to the dedicated to queue the next map.
		self._previous_map = None
		self._current_map = None
		self._next_map = None

		# Hold the original TA limit for the /extend and //extend functions.
		self._is_extended = False
		self._original_ta = None

		# Regular Expression to extract the MX-ID from a filename.
		self._mx_id_regex = re.compile('(?:PyPlanet-MX\\/)([A-Z]{2})-(\\d+)\\.')

	async def on_start(self):
		self._instance.signals.listen('maniaplanet:playlist_modified', lambda: '')
		self._instance.signals.listen('maniaplanet:podium_start', self._podium_start)

		# Fully update list + database.
		await self.update_list(full_update=True)

		# Get current and next map.
		self._current_map, self._next_map = await asyncio.gather(
			self.handle_map_change(await self._instance.gbx('GetCurrentMapInfo')),
			self.handle_map_change(await self._instance.gbx('GetNextMapInfo')),
		)
		self._previous_map = None

	async def handle_map_change(self, info):
		"""
		This will be called from the glue that creates the signal 'maniaplanet:map_begin' or 'map_end'.

		:param info: Mapinfo in dict.
		:return: Map instance.
		:rtype: pyplanet.apps.core.maniaplanet.models.map.Map
		"""
		# Try to retrieve the MX-id from the filename.
		mx_id = self._extract_mx_id(info['FileName'])

		# Get or create.
		map_info = await Map.get_or_create_from_info(
			uid=info['UId'], name=info['Name'], author_login=info['Author'], file=info['FileName'],
			environment=info['Environnement'], map_type=info['MapType'], map_style=info['MapStyle'],
			num_laps=info['NbLaps'], num_checkpoints=info['NbCheckpoints'], time_author=info['AuthorTime'],
			time_bronze=info['BronzeTime'], time_silver=info['SilverTime'], time_gold=info['GoldTime'],
			price=info['CopperPrice'], mx_id=mx_id,
		)
		self._previous_map = self._current_map
		self._current_map = map_info
		return map_info

	async def handle_playlist_change(self, source, **kwargs):
		pass
		# if source and source[2]:
		# 	return await self.update_list(full_update=True)

	def _extract_mx_id(self, file_name):
		"""
		Try to extract the MX-id from a filename.

		:param file_name: File name from Dedicated.
		:type file_name: str
		:return: String or None
		"""
		matches = re.findall(self._mx_id_regex, file_name)
		if not matches or len(matches) != 1 or len(matches[0]) != 2:
			return None
		return matches[0][1]

	async def _podium_start(self, **kwargs):
		"""
		Handle start of podium to reset ta limit if extended.

		:param kwargs:
		:return:
		"""
		# Set back the timer if time has been extended.
		if self._is_extended and self._original_ta:
			mode_settings = await self._instance.mode_manager.get_settings()
			if 'S_TimeLimit' not in mode_settings:
				return

			mode_settings['S_TimeLimit'] = self._original_ta
			await self._instance.mode_manager.update_settings(mode_settings)

			self._is_extended = False
			self._original_ta = None

	async def update_list(self, full_update=False, detach_fks=True):
		raw_list = await self._instance.gbx('GetMapList', -1, 0)
		updated = list()

		if full_update:
			# Query all existing entries from database.
			maps = list(await Map.execute(
				Map.select().where(Map.uid << [m['UId'] for m in raw_list])
			))

			db_uids = [m.uid for m in maps]
			diff = [x for x in raw_list if x['UId'] not in db_uids]

			# Insert all missing maps into the DB.
			rows = list()
			for details in diff:
				mx_id = self._extract_mx_id(details['FileName'])

				# HACK: Due to a limited map name length of 150 chars, we want to strip it to the maximum possible.
				# This is a temporary fix and should be better handled in the future.
				name = details['Name']
				if len(name) > 150:
					name = name[:150]
					logging.getLogger(__name__).warning('Map name is very long, truncating to 150 chars.')

				rows.append(dict(
					uid=details['UId'], file=details['FileName'], name=name, author_login=details['Author'],
					environment=details['Environnement'], time_gold=details['GoldTime'], price=details['CopperPrice'],
					map_type=details['MapType'], map_style=details['MapStyle'], mx_id=mx_id
				))

			if len(rows) > 0:
				await Map.execute(Map.insert_many(rows))
				maps += list(await Map.execute(
					Map.select().where(Map.uid << [m['uid'] for m in rows])
				))

			async with self.lock:
				self._maps = set(maps)

			# Reload locals for all maps.
			# TODO: Find better way to remove this and handle it on the folders way.
			coroutines = list()
			if 'local_records' in self._instance.apps.apps:
				if detach_fks:
					asyncio.ensure_future(self._instance.apps.apps['local_records'].load_map_locals())
				else:
					coroutines.append(self._instance.apps.apps['local_records'].load_map_locals())

			# Reload karma for all maps.
			if 'karma' in self._instance.apps.apps:
				if detach_fks:
					asyncio.ensure_future(self._instance.apps.apps['karma'].load_map_votes())
				else:
					coroutines.append(self._instance.apps.apps['karma'].load_map_votes())

			if coroutines:
				await asyncio.gather(*coroutines)
		else:
			# Only update/insert the changed bits, (not checking for removed maps!!).
			async with self.lock:
				for details in raw_list:
					if not any(m.uid == details['UId'] for m in self._maps):
						# Detect any MX-id from the filename.
						mx_id = self._extract_mx_id(details['FileName'])

						# Map not yet in self._maps. Add it.
						map_instance = await Map.get_or_create_from_info(
							details['UId'], details['FileName'], details['Name'], details['Author'],
							environment=details['Environnement'], time_gold=details['GoldTime'],
							price=details['CopperPrice'], map_type=details['MapType'], map_style=details['MapStyle'],
							mx_id=mx_id,
						)
						self._maps.add(map_instance)
						updated.append(map_instance)
		return updated

	async def get_map(self, uid=None):
		"""
		Get map instance by uid.

		:param uid: By uid (pk).
		:return: Player or exception if not found
		"""
		try:
			return await Map.get_by_uid(uid)
		except DoesNotExist:
			raise MapNotFound('Map not found.')

	async def get_map_by_index(self, index):
		"""
		Get map instance by index id (primary key).

		:param index: Primary key index.
		:return: Map instance or raise exception.
		"""
		try:
			return await Map.get(id=index)
		except DoesNotExist:
			raise MapNotFound('Map not found.')

	@property
	def next_map(self):
		"""
		The next scheduled map.

		:rtype: pyplanet.apps.core.maniaplanet.models.Map
		"""
		return self._next_map

	async def set_next_map(self, map):
		"""
		Set the next map. This will prepare the manager to set the next map and will communicate the next map to the
		dedicated server.

		The Map parameter can be a map instance or the UID of the map.

		:param map: Map instance or UID string.
		:type map: pyplanet.apps.core.maniaplanet.models.Map, str
		"""
		if isinstance(map, str):
			map = await self.get_map(map)
		if not isinstance(map, Map):
			raise Exception('When setting the map, you should give an Map instance!')
		if map.file:
			await self._instance.gbx('ChooseNextMap', map.file)
		else:
			await self._instance.gbx('SetNextMapIdent', map.uid)
		self._next_map = map

	@property
	def current_map(self):
		"""
		The current map, database model instance.

		:rtype: pyplanet.apps.core.maniaplanet.models.Map
		"""
		return self._current_map

	@property
	def previous_map(self):
		"""
		The previously played map, or None if not known!

		:rtype: pyplanet.apps.core.maniaplanet.models.Map
		"""
		return self._previous_map

	@property
	def maps(self):
		"""
		Get the maps that are currently loaded on the server. The list should contain model instances of the currently
		loaded matchsettings. This list should be up-to-date.

		:rtype: list
		"""
		return self._maps

	async def set_current_map(self, map):
		"""
		Set the current map and jump to it.

		:param map: Map instance or uid.
		"""
		if isinstance(map, str):
			map = await self.get_map(map)
		if not isinstance(map, Map):
			raise Exception('When setting the map, you should give an Map instance!')

		await self._instance.gbx('JumpToMapIdent', map.uid)
		self._next_map = map

	def playlist_has_map(self, uid):
		"""
		Check if our current playlist has a map with the UID given.

		:param uid: UID String
		:return: Boolean, True if it's in our current playlist (match settings in our session).
		"""
		for map_instance in self.maps:
			if map_instance.uid == uid:
				return True
		return False

	async def add_map(self, filename, insert=True, save_matchsettings=True):
		"""
		Add or insert map to current online playlist.

		:param filename: Load from filename relative to the 'Maps' directory on the dedicated host server.
		:param insert: Insert after the current map, this will make it play directly after the current map. True by default.
		:param save_matchsettings: Save match settings as well.
		:type filename: str
		:type insert: bool
		:type save_matchsettings: bool
		:raise: pyplanet.contrib.map.exceptions.MapIncompatible
		:raise: pyplanet.contrib.map.exceptions.MapException
		"""
		gbx_method = 'InsertMap' if insert else 'AddMap'

		try:
			result = await self._instance.gbx(gbx_method, filename)
		except Fault as e:
			if 'unknown' in e.faultString:
				raise MapNotFound('Map is not found on the server.')
			elif 'already' in e.faultString:
				raise MapException('Map already added to server.')
			raise MapException(e.faultString)

		# Try to save match settings.
		try:
			if save_matchsettings:
				await self.save_matchsettings()
		except Exception as e:
			handle_exception(e, __name__, 'add_map', extra_data={'EXTRAHOOK': 'Map Insert bug, see #306'})

		return result

	async def upload_map(self, fh, filename, insert=True, overwrite=False):
		"""
		Upload and add/insert the map to the current online playlist.

		:param fh: File handler, bytesio object or any readable context.
		:param filename: The filename when saving on the server. Must include the map.gbx! Relative to 'Maps' folder.
		:param insert: Insert after the current map, this will make it play directly after the current map. True by default.
		:param overwrite: Overwrite current file if exists? Default False.
		:type filename: str
		:type insert: bool
		:type overwrite: bool
		:raise: pyplanet.contrib.map.exceptions.MapIncompatible
		:raise: pyplanet.contrib.map.exceptions.MapException
		:raise: pyplanet.core.storage.exceptions.StorageException
		"""
		exists = await self._instance.storage.driver.exists(filename)
		if exists and not overwrite:
			raise MapException('Map with filename already located on server!')
		if not exists:
			await self._instance.storage.driver.touch('{}{}'.format(self._instance.storage.MAP_FOLDER, filename))

		async with self._instance.storage.open_map(filename, 'wb+') as fw:
			await fw.write(fh.read(-1))

		return await self.add_map(filename, insert=insert)

	async def remove_map(self, map, delete_file=False):
		"""
		Remove and optionally delete file from server and playlist.

		:param map: Map instance or filename in string.
		:param delete_file: Boolean to decide if we are going to remove the file from the server too. Defaults to False.
		:type delete_file: bool
		:raise: pyplanet.contrib.map.exceptions.MapException
		:raise: pyplanet.core.storage.exceptions.StorageException
		"""
		if isinstance(map, Map):
			map = map.file
		if not isinstance(map, str):
			raise ValueError('Map must be instance or string uid!')

		try:
			success = await self._instance.gbx('RemoveMap', map)
			if success:
				the_map = None
				for m in self._maps:
					if m.file == map:
						the_map = m
						break
				if the_map:
					self._maps.remove(the_map)
		except Fault as e:
			if 'unknown' in e.faultString:
				raise MapNotFound('Dedicated can\'t find map. Already removed?')
			raise MapException('Error when removing map from playlist: {}'.format(e.faultString))

		# Try to save match settings.
		try:
			await self.save_matchsettings()
		except:
			pass

		# Delete the actual file.
		if delete_file:
			try:
				await self._instance.storage.remove_map(map)
			except:
				raise MapException('Can\'t delete map file after removing from playlist.')

	async def _override_timelimit(self, filename):
		"""
		Called to overwrite S_TimeLimit in MatchSettings file if the current map is extended
		
		:param filename: Give the filename of the matchsettings.
		"""
		if self._is_extended and self._original_ta:
			try:
				async with self._instance.storage.open(os.path.join(self._instance.game.server_map_dir, filename), 'r+') as f:
					content = await f.readlines()
					for i in range(len(content)):
						if 'S_TimeLimit' in content[i]:
							content[i] = re.sub('value="(.+?)"', 'value="{}"'.format(self._original_ta), content[i])
							await f.seek(0)
							await f.write(''.join(content))
							await f.truncate()
							break
			except:
				logging.getLogger(__name__).warning('Can\'t update matchsettings with original time limit to \'{}\'!'.format(filename))

	async def save_matchsettings(self, filename=None):
		"""
		Save the current playlist and configuration to the matchsettings file.

		:param filename: Give the filename of the matchsettings, Leave empty to use the current loaded and configured one.
		:type filename: str
		:raise: pyplanet.contrib.map.exceptions.MapException
		:raise: pyplanet.core.storage.exceptions.StorageException
		"""
		setting = settings.MAP_MATCHSETTINGS
		if isinstance(setting, dict) and self._instance.process_name in setting:
			setting = setting[self._instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not filename and not setting:
			raise ImproperlyConfigured(
				'The setting \'MAP_MATCHSETTINGS\' is not configured for this server! We can\'t save the Match Settings!'
			)
		if not filename:
			filename = 'MatchSettings/{}'.format(
				setting.format(server_login=self._instance.game.server_player_login)
			)

		try:
			await self._instance.gbx('SaveMatchSettings', filename)
			await self._override_timelimit(filename)
		except Exception as e:
			logging.exception(e)
			raise MapException('Can\'t save matchsettings to \'{}\'!'.format(filename)) from e

	async def load_matchsettings(self, filename):
		"""
		Load Match Settings file and insert it into the current map playlist.

		:param filename: File to load, relative to Maps folder.
		:return: Boolean if loaded.
		"""
		try:
			if not await self._instance.storage.driver.exists(
				os.path.join(self._instance.storage.MAP_FOLDER, filename)
			):
				raise MapException('Can\'t find match settings file. Does it exist?')
			else:
				await self._instance.gbx('LoadMatchSettings', filename)
		except Exception as e:
			logging.warning('Can\'t load match settings!')
			raise MapException('Can\'t load matchsettings according the dedicated server, tried loading from \'{}\'!'.format(filename)) from e

	async def extend_ta(self, extend_with=None):
		"""
		Extend time limit of the current map.
		Extend with given seconds, or double the original TA timer if None is given.

		:param extend_with: Extend with the given seconds, or None for adding the original TA limit to the current limit(double)
		:type extend_with: int
		:return:
		"""
		mode_settings = await self._instance.mode_manager.get_settings()
		if 'S_TimeLimit' not in mode_settings:
			raise ModeIncompatible('Current mode doesn\'t support the extend TA method. Not Time Attack?')

		temp_mode_settings = mode_settings.copy()
		original_ta = self._original_ta or temp_mode_settings['S_TimeLimit']

		if not extend_with:
			extend_with = original_ta
		if extend_with > 2000000000:
			extend_with = 2000000000

		temp_mode_settings['S_TimeLimit'] += abs(extend_with)
		if temp_mode_settings['S_TimeLimit'] > 2000000000:
			temp_mode_settings['S_TimeLimit'] = 2000000000

		if not self._is_extended or not self._original_ta:
			self._original_ta = mode_settings['S_TimeLimit']
		self._is_extended = True

		await self._instance.mode_manager.update_settings(temp_mode_settings)
		return extend_with



from pyplanet.utils import toposort


def test_toposort():
	dep_tree = {
		'node_root': [],
		'node_1': ['node_root', 'node_3', 'node_2'],
		'node_2': ['node_root'],
		'node_3': ['node_2']
	}
	order = toposort.toposort(dep_tree)
	assert order == ['node_root', 'node_2', 'node_3', 'node_1']

	dep_tree = {
		'node_root': [],
		'node_1': ['node_root', 'node_3', 'node_2'],
		'node_2': ['node_root'],
		'node_3': ['node_2', 'node_1']
	}
	cyclical = False
	try:
		toposort.toposort(dep_tree)
	except ValueError as e:
		if 'graph is cyclical through' in str(e):
			cyclical = True

	assert cyclical is True



import asyncio
import json
import uuid
from xmlrpc.client import dumps

from pyplanet.core.exceptions import TransportException


class Query:
	def __init__(self, client, method, *args, timeout=45, **kwargs):
		"""
		Initiate the prepared query.

		:param client: GbxClient to execute actions.
		:param method: Method name
		:param args: Arguments
		:param timeout: Timeout of the call.
		:type client: pyplanet.core.gbx.client.GbxClient
		:type method: str:
		"""
		self._client = client

		self.packet = None
		self.length = None

		self.method = method
		self.args = args
		self.timeout = timeout
		self.result = None

	async def execute(self):
		"""
		Execute call.

		:return: Future with results.
		:rtype: Future<any>
		"""
		return await self._client.execute(self.method, *self.args, timeout=self.timeout)

	def __await__(self):
		"""
		Execute query directly.
		"""
		return self.execute().__await__()

	def prepare(self):
		"""
		Prepare the query, marshall the payload, create binary data and calculate length (size).
		"""
		self.packet = dumps(self.args, methodname=self.method, allow_none=True).encode()
		self.length = len(self.packet)

		if (self.length + 8) > self._client.MAX_REQUEST_SIZE:
			raise TransportException('The prepared query is larger than the maximum request size, we will not send this query!')

	async def __aenter__(self):
		self.result = await self.execute()
		return self

	async def __aexit__(self, exc_type, exc_val, exc_tb):
		self.result = None


class ScriptQuery(Query):

	def __init__(self, client, method, *args, timeout=15, encode_json=True, response_id=True):
		"""
		Initiate a Scripted Query.

		:param client: Client instance
		:param method: Method name
		:param args: Arguments
		:param timeout: Timeout to wait for future result.
		:param encode_json: Is body json? True by default.
		:param response_id: Is request requiring response_id?
		:type client: pyplanet.core.gbx.client.GbxClient
		"""
		# Make sure we call the script stuff with TriggerModeScriptEventArray.
		gbx_method = 'TriggerModeScriptEventArray'
		gbx_args = list()

		# Make sure we generate a response_id.
		if response_id is True:
			self.response_id = uuid.uuid4().hex
		else:
			self.response_id = None

		# Encode to json if args are given, and encode_json is true (default).
		if encode_json and len(args) > 0:
			gbx_args.append(json.dumps(args))
		elif not encode_json and len(args) > 0:
			gbx_args.extend(args)

		# Add the response_id to the end of the argument list.
		if self.response_id:
			gbx_args.append(str(self.response_id))

		super().__init__(client, gbx_method, method, gbx_args, timeout=timeout)

	async def execute(self):
		"""
		Execute call.

		:return: Future with results.
		:rtype: Future<any>
		"""
		# Create new future to be returned. This future will be the answered inside a script callback.
		future = None
		if self.response_id:
			self._client.script_handlers[self.response_id] = future = asyncio.Future()

		# Execute the call itself and register the callback script handler.
		gbx_res = await self._client.execute(self.method, *self.args)

		if self.response_id:
			return await asyncio.wait_for(future, self.timeout)  # Timeout after 15 seconds!
		return gbx_res



import asyncio
import gzip
import logging
import requests

from xmlrpc.client import dumps, loads

from requests import ConnectTimeout, ReadTimeout

from pyplanet import __version__ as version
from pyplanet.apps.contrib.dedimania.exceptions import DedimaniaTransportException, DedimaniaFault, \
	DedimaniaNotSupportedException, DedimaniaInvalidCredentials
from pyplanet.utils.log import handle_exception

logger = logging.getLogger(__name__)


class DedimaniaAPI:
	API_URL = 'http://dedimania.net:8082/Dedimania'

	def __init__(self, instance, server_login, dedi_code, path, pack_mask, server_version, server_build, game='TM2'):
		"""
		Initiate dedi api.

		:param instance: ControllerInstance
		:param server_login: .
		:param dedi_code: .
		:param path: .
		:param pack_mask: .
		:param server_version: .
		:param server_build: .
		:param game: Game info
		:type instance: pyplanet.core.instance.Instance
		"""
		self.instance = instance
		self.loop = instance.loop
		self.client = requests.session()
		self.headers = {
			'User-Agent': 'PyPlanet/{}'.format(version),
			'Accept': 'text/xml',
			'Accept-Encoding': 'gzip',
			'Content-Type': 'text/xml; charset=UTF-8',
			'Content-Encoding': 'gzip',
			'Keep-Alive': 'timeout=600, max=2000',
			'Connection': 'Keep-Alive',
		}

		self.server_login = server_login
		self.dedimania_code = dedi_code
		self.path = path
		self.pack_mask = pack_mask
		self.server_version = server_version
		self.server_build = server_build
		self.game = game

		self.update_task = None
		self.session_id = None
		self.retries = 0

	async def on_start(self):
		pass

	def mode_to_dedi_mode(self, mode):
		mode = mode.lower()
		if mode.startswith('teamattack') or mode.startswith('chase'):
			return False
		elif mode.startswith('rounds') or mode.startswith('team') or mode.startswith('cup'):
			return 'Rounds'
		elif mode.startswith('timeattack') or mode.startswith('laps') or mode.startswith('doppler'):
			return 'TA'
		return False

	def __request(self, body):
		return self.client.request('POST', self.API_URL, None, body, self.headers, timeout=10)

	async def execute(self, method, *args):
		payload = dumps(args, methodname=method, allow_none=True)
		body = gzip.compress(payload.encode('utf8'))
		try:
			res = await self.loop.run_in_executor(None, self.__request, body)
			data, _ = loads(res.text, use_datetime=True)
			if isinstance(data, (tuple, list)) and len(data) > 0 and len(data[0]) > 0:
				if isinstance(data[0][0], dict) and 'faultCode' in data[0][0]:
					raise DedimaniaFault(faultCode=data[0][0]['faultCode'], faultString=data[0][0]['faultString'])
				self.retries = 0
				return data[0]
			raise DedimaniaTransportException('Invalid response from dedimania!')
		except (ConnectionError, ReadTimeout, ConnectionRefusedError, requests.exceptions.ConnectionError) as e:
			raise DedimaniaTransportException(e) from e
		except ConnectTimeout as e:
			raise DedimaniaTransportException(e) from e
		except DedimaniaTransportException:
			# Try to setup new session.
			self.retries += 1
			if self.retries > 5:
				raise DedimaniaTransportException('Dedimania didn\'t gave the right answer after few retries!')
			self.client = requests.session()
			try:
				await self.authenticate()
				return await self.execute(method, *args)
			except Exception as e:
				logger.error('XML-RPC Fault retrieved from Dedimania: {}'.format(str(e)))
				handle_exception(e, __name__, 'execute')
				raise DedimaniaTransportException('Could not retrieve data from dedimania!')
		except DedimaniaFault as e:
			if 'Bad SessionId' in e.faultString or ('SessionId' in e.faultString and 'not found' in e.faultString):
				try:
					self.retries += 1
					if self.retries > 5:
						raise DedimaniaTransportException('Max retries reached for reauthenticating with dedimania!')

					# Save original session ID.
					original_session_id = '{}'.format(self.session_id)

					# Reauthenticate
					await self.authenticate()

					# Replace session_id in args.
					if len(args) > 0 and len(args[0]) > 0 and isinstance(args[0][0], dict) and 'params' in args[0][0]:
						new_params = list(args[0][0]['params'])
						if len(new_params) > 0 and isinstance(new_params[0], str) and new_params[0] == original_session_id:
							new_params[0] = self.session_id
							args[0][0]['params'] = tuple(new_params)

					# Try again.
					return await self.execute(method, *args)
				except:
					return
			logger.error('XML-RPC Fault retrieved from Dedimania: {}'.format(str(e)))
			handle_exception(e, __name__, 'execute', extra_data={
				'dedimania_retries': self.retries,
			})
			raise DedimaniaTransportException('Could not retrieve data from dedimania!')

	async def multicall(self, *queries):
		queries = queries + (
			(
				'dedimania.WarningsAndTTR',
			),
		)
		return await self.execute('system.multicall', [{'methodName': c[0], 'params': c[1:]} for c in queries])

	async def authenticate(self):
		try:
			result = await self.multicall(
				('dedimania.OpenSession', {
					'Game': self.game, 'Login': self.server_login, 'Code': self.dedimania_code, 'Path': self.path,
					'Packmask': self.pack_mask, 'ServerVersion': self.server_version, 'ServerBuild': self.server_build,
					'Tool': 'PyPlanet', 'Version': str(version)
				})
			)
		except DedimaniaTransportException as e:
			logger.error('Dedimania Error during authentication: {}'.format(str(e)))
			return
		if not result:
			logger.error('Dedimania response doesn\'t contain authentication results!')
			return

		try:
			if 'Error' in result[0][0] and 'Bad code' in result[0][0]['Error'].lower():
				raise DedimaniaInvalidCredentials('Bad code or login!')
		except DedimaniaInvalidCredentials:
			raise
		except:
			pass

		self.session_id = result[0][0]['SessionId']

		if not self.update_task:
			self.update_task = asyncio.ensure_future(self.update_loop())

		return self.session_id

	async def update_loop(self):
		while True:
			await asyncio.sleep(60 * 4)

			if not self.session_id:
				continue

			def is_spectator(player):
				return bool(player['SpectatorStatus'] % 10)

			players = await self.instance.gbx('GetPlayerList', -1, 0)
			player_list = [
				{'Login': p['Login'], 'IsSpec': is_spectator(p), 'Vote': -1} for p in players if p['Login'] != self.instance.game.server_player_login
			]

			num_specs = sum(p['IsSpec'] for p in player_list)
			num_players = len(player_list) - num_specs
			max_players = self.instance.player_manager.max_players
			max_specs = self.instance.player_manager.max_spectators
			mode = self.mode_to_dedi_mode(await self.instance.mode_manager.get_current_script())

			if not mode:
				continue
			try:
				response = await self.multicall(
					('dedimania.UpdateServerPlayers', self.session_id, {
						'SrvName': self.instance.game.server_name, 'Comment': '', 'Private': self.instance.game.server_is_private, 'NumPlayers': num_players,
						'MaxPlayers': max_players, 'NumSpecs': num_specs, 'MaxSpecs': max_specs
					}, {
						 'UId': self.instance.map_manager.current_map.uid, 'GameMode': mode,
					 }, player_list)
				)
				assert response[0][0] is True
			except DedimaniaTransportException as e:
				logger.warning('Dedimania update call failed! {}'.format(str(e)))

	async def player_connect(
		self, login, nickname, path, is_spec
	):
		if not self.session_id:
			return None
		try:
			response = await self.multicall(
				('dedimania.PlayerConnect', self.session_id, login, nickname, path, is_spec)
			)
			if not response:
				return None
			response = response[0][0]
			return dict(
				banned=bool(response['Banned']), login=response['Login'], max_rank=response['MaxRank'],
			)
		except DedimaniaTransportException as e:
			logging.exception(e)
			return None

	async def player_disconnect(self, login, tool_option):
		if not self.session_id:
			return True
		await self.multicall(
			('dedimania.PlayerDisconnect', self.session_id, login, tool_option)
		)
		return True

	async def get_map_details(
		self, map, game_mode, server_name, server_comment, is_private, max_players, max_specs, players, server_login
	):
		"""
		Get records for specific map instance.

		:param map: Map instance.
		:param game_mode: Game mode, either 'Rounds' or 'TA'.
		:param server_name: Name of server
		:param server_comment: Comment
		:param is_private: Is server hidden from lists?
		:param max_players: Maximum number of players.
		:param max_specs: Maximum of spectators.
		:param players: List with players, raw from gbx command!!!!!
		:type map: pyplanet.apps.core.maniaplanet.models.map.Map
		:return: Record list.
		"""
		if not self.session_id:
			await self.authenticate()
		if not self.session_id:
			raise DedimaniaTransportException('Dedimania not authenticated!')

		def is_spectator(player):
			return bool(player['SpectatorStatus'] % 10)

		player_list = [
			{'Login': p['Login'], 'IsSpec': is_spectator(p)} for p in players if p['Login'] != server_login
		]
		num_specs = sum(p['IsSpec'] for p in player_list)
		num_players = len(player_list) - num_specs

		mode = self.mode_to_dedi_mode(game_mode)
		if not mode:
			raise DedimaniaNotSupportedException('Mode is not supported!')

		result = await self.multicall(
			('dedimania.GetChallengeRecords', self.session_id, {
				'UId': map.uid, 'Name': map.name, 'Environment': map.environment, 'Author': map.author_login,
				'NbCheckpoints': map.num_checkpoints, 'NbLaps': map.num_laps,
			}, mode, {
				'SrvName': server_name, 'Comment': server_comment, 'Private': is_private, 'NumPlayers': num_players,
				'MaxPlayers': max_players, 'NumSpecs': num_specs, 'MaxSpecs': max_specs
			}, player_list)
		)
		if not result or not isinstance(result, list):
			raise DedimaniaTransportException('Result seems to be empty!')

		result = result[0][0]
		if not result:
			raise DedimaniaTransportException('Result seems to be empty!')

		allowed_modes = result['AllowedGameModes']
		server_max_rank = result['ServerMaxRank']
		response_players = result['Players']
		raw_records = result['Records']
		records = [
			DedimaniaRecord(
				r['Login'],
				r['NickName'],
				r['Best'],
				r['Rank'],
				r['MaxRank'],
				[int(c) for c in r['Checks'].split(',')],
				r['Vote']
			)
			for r in raw_records
		]
		return server_max_rank, allowed_modes, response_players, records or []

	async def set_map_times(self, map, game_mode, records, v_replay=None, v_replay_checks=None, ghost_replay=None):
		mode = self.mode_to_dedi_mode(game_mode)
		if not mode:
			raise DedimaniaNotSupportedException('Mode is not supported!')

		if not self.session_id:
			raise DedimaniaTransportException('Dedimania not authenticated!')

		times = [{
			'Login': r.login, 'Best': r.score, 'Checks': ','.join([str(c) for c in r.cps]),
		} for r in records if r.updated]
		replays = {
			'VReplay': v_replay or None,
			'VReplayChecks': v_replay_checks or '',
			'Top1GReplay': ghost_replay or ''
		}
		if not replays['VReplay']:
			# Nothing to update!
			logger.debug('Dedimania end map, nothing new to update! Skipping set times call!')
			return

		result = await self.multicall(
			('dedimania.SetChallengeTimes', self.session_id, {
				'UId': map.uid, 'Name': map.name, 'Environment': map.environment, 'Author': map.author_login,
				'NbCheckpoints': map.num_checkpoints, 'NbLaps': map.num_laps,
			}, mode, times, replays)
		)

		try:
			return bool(isinstance(result[0][0]['Records'], list))
		except Exception as e:
			logger.error('Sending times to dedimania failed. Info: {}'.format(result))
			return False


class DedimaniaRecord:

	def __init__(
		self, login, nickname, score, rank, max_rank, cps, vote
	):
		self.login = login
		self.nickname = nickname
		self.score = score
		self.rank = rank
		self.max_rank = max_rank
		self.cps = cps
		self.vote = vote

		self.race_cps = list()

		self.updated = False
		self.new_index = None
